from typing import Literal
from typing import Union

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class MwsPermissionAssignment(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Mws Permission Assignment

    Attributes
    ----------
    permissions:
        The list of workspace permissions to assign to the principal:
            - "USER" Can access the workspace with basic privileges.
            - "ADMIN" Can access the workspace and has workspace admin
              privileges to manage users and groups, workspace configurations,
              and more.
    principal_id:
        Databricks ID of the user, service principal, or group. The principal
        ID can be retrieved using the SCIM API, or using databricks_user,
        databricks.ServicePrincipal or databricks.Group data sources.
    workspace_id:
        Databricks workspace ID.

    Examples
    --------
    ```py
    ```
    """

    permissions: list[Literal["USER", "ADMIN"]] = ["USER"]  # required
    principal_id: Union[int, str] = None  # required
    workspace_id: Union[int, str] = None  # required

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_type_id(self):
        return "permission"

    @property
    def resource_key(self):
        return f"{self.principal_id}-{self.workspace_id}"

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:MwsPermissionAssignment"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_mws_permission_assignment"



import os
import shutil
import uuid
import warnings
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Literal
from typing import Union

from pydantic import field_validator
from pydantic import model_validator

from laktory._logger import get_logger
from laktory._settings import settings
from laktory.exceptions import DataQualityExpectationsNotSupported
from laktory.models.basemodel import BaseModel
from laktory.models.dataquality.expectation import DataQualityExpectation
from laktory.models.datasinks import DataSinksUnion
from laktory.models.datasinks import TableDataSink
from laktory.models.datasources import BaseDataSource
from laktory.models.datasources import DataSourcesUnion
from laktory.models.datasources import PipelineNodeDataSource
from laktory.models.datasources import TableDataSource
from laktory.models.pipeline.pipelinechild import PipelineChild
from laktory.models.transformers.polarschain import PolarsChain
from laktory.models.transformers.polarschainnode import PolarsChainNode
from laktory.models.transformers.sparkchain import SparkChain
from laktory.models.transformers.sparkchainnode import SparkChainNode
from laktory.spark import SparkSession
from laktory.typing import AnyDataFrame

logger = get_logger(__name__)


class PipelineNode(BaseModel, PipelineChild):
    """
    Pipeline base component generating a DataFrame by reading a data source and
    applying a transformer (chain of dataframe transformations). Optional
    output to a data sink. Some basic transformations are also natively
    supported.

    Attributes
    ----------
    add_layer_columns:
        If `True` and `layer` not `None` layer-specific columns like timestamps
        are added to the resulting DataFrame.
    dlt_template:
        Specify which template (notebook) to use if pipeline is run with
        Databricks Delta Live Tables. If `None` default laktory template
        notebook is used.
    drop_duplicates:
        If `True`:
            - drop duplicated rows using `primary_key` if defined or all
              columns if not defined.
        If list of strings:
            - drop duplicated rows using `drop_duplicates` as the subset.
    drop_source_columns:
        If `True`, drop columns from the source after read and only keep
        columns resulting from executing the SparkChain.
    expectations:
        List of expectations for the DataFrame. Can be used as warnings, drop
        invalid records or fail a pipeline.
    layer:
        Layer in the medallion architecture
    name:
        Name given to the node. Required to reference a node in a data source.
    primary_keys:
        A list of column names that uniquely identify each row in the
        DataFrame. These columns are used to:
            - Document the uniqueness constraints of the node's output data.
            - Define the default subset for dropping duplicate rows if no
              explicit subset is provided in `drop_duplicates`.
            - Define the default primary keys for sinks CDC merge operations
            - Referenced in expectations and unit tests.
        While optional, specifying `primary_keys` helps enforce data integrity
        and ensures that downstream operations, such as deduplication, are
        consistent and reliable.
    root_path:
        Location of the pipeline node root used to store logs, metrics and
        checkpoints.
    source:
        Definition of the data source
    sinks:
        Definition of the data sink(s). Set `is_quarantine` to True to store
        node quarantine DataFrame.
    transformer:
        Spark or Polars chain defining the data transformations applied to the
        data source.
    timestamp_key:
        Name of the column storing a timestamp associated with each row. It is
        used as the default column by the builder when creating watermarks.


    Examples
    --------
    A node reading stock prices data from a CSV file and writing a DataFrame
    to disk as a parquet file.
    ```py
    import io

    from laktory import models

    node_yaml = '''
        name: brz_stock_prices
        layer: BRONZE
        source:
            format: CSV
            path: ./raw/brz_stock_prices.csv
        sinks:
        -   format: PARQUET
            mode: OVERWRITE
            path: ./dataframes/brz_stock_prices
    '''

    node = models.PipelineNode.model_validate_yaml(io.StringIO(node_yaml))

    # node.execute(spark)
    ```

    A node reading stock prices from an upstream node and writing a DataFrame
    to a data table.
    ```py
    import io

    from laktory import models

    node_yaml = '''
        name: slv_stock_prices
        layer: SILVER
        source:
          node_name: brz_stock_prices
        sinks:
        - catalog_name: hive_metastore
          schema_name: default
          table_name: slv_stock_prices
        transformer:
          nodes:
          - with_column:
              name: created_at
              type: timestamp
              expr: data.created_at
          - with_column:
              name: symbol
              expr: data.symbol
          - with_column:
              name: close
              type: double
              expr: data.close
    '''

    node = models.PipelineNode.model_validate_yaml(io.StringIO(node_yaml))

    # node.execute(spark)
    ```
    """

    add_layer_columns: bool = True
    dlt_template: Union[str, None] = "DEFAULT"
    dataframe_backend: Literal["SPARK", "POLARS"] = None
    description: str = None
    drop_duplicates: Union[bool, list[str], None] = None  # TODO: Migrate to transformer
    drop_source_columns: Union[bool, None] = None  # TODO: Migrate to transformer
    transformer: Union[SparkChain, PolarsChain, None] = None
    expectations: list[DataQualityExpectation] = []
    expectations_checkpoint_location: str = None
    layer: Literal["BRONZE", "SILVER", "GOLD"] = None
    name: Union[str, None] = None
    primary_keys: list[str] = None
    sinks: list[DataSinksUnion] = None
    root_path: str = None
    source: DataSourcesUnion
    timestamp_key: str = None
    _view_definition: str = None
    _stage_df: Any = None
    _output_df: Any = None
    _quarantine_df: Any = None
    _source_columns: list[str] = []

    @model_validator(mode="before")
    @classmethod
    def push_df_backend(cls, data: Any) -> Any:
        """Need to push dataframe_backend which is required to differentiate between spark and polars transformer"""
        df_backend = data.get("dataframe_backend", None)
        if df_backend:
            for k in ["source", "transformer"]:
                o = data.get(k, None)
                if o and isinstance(o, dict):
                    # source or transformer as a dict
                    o["dataframe_backend"] = o.get("dataframe_backend", df_backend)
                elif o:
                    # source or transformer as a model
                    o.dataframe_backend = o.dataframe_backend or df_backend

        return data

    @field_validator("root_path", "expectations_checkpoint_location", mode="before")
    @classmethod
    def posixpath_to_string(cls, value: Any) -> Any:
        if isinstance(value, Path):
            value = str(value)
        return value

    @model_validator(mode="after")
    def default_values(self) -> Any:
        # Default values
        if self.layer == "BRONZE":
            if self.drop_source_columns is None:
                self.drop_source_columns = False
            if self.drop_duplicates is not None:
                self.drop_duplicates = False

        if self.layer == "SILVER":
            if self.drop_source_columns is None:
                self.drop_source_columns = True
            if self.drop_duplicates is not None and self.primary_keys:
                self.drop_duplicates = True

        #
        # if self.layer == "GOLD":
        #     if self.drop_source_columns is None:
        #         self.drop_source_columns = False
        #     if self.drop_duplicates is not None:
        #         self.drop_duplicates = False
        #

        # Generate node id
        if self.name is None:
            self.name = str(uuid.uuid4())

        return self

    @model_validator(mode="after")
    def push_primary_keys(self) -> Any:
        # Assign primary keys
        if self.primary_keys and self.sinks:
            for s in self.sinks:
                if s.is_cdc:
                    if s.merge_cdc_options.primary_keys is None:
                        s.merge_cdc_options.primary_keys = self.primary_keys

        return self

    @model_validator(mode="after")
    def validate_sinks(self):
        if self.has_output_sinks:
            count = 0
            for s in self.output_sinks:
                if s.is_primary:
                    count += 1
            if count != 1:
                raise ValueError(
                    f"Node '{self.name}' must have exactly one primary sink. Currently have {count}"
                )
        return self

    @model_validator(mode="after")
    def validate_expectations(self):
        if self.source.as_stream:
            # Expectations type
            for e in self.expectations:
                if not e.is_streaming_compatible:
                    raise DataQualityExpectationsNotSupported(e, self)

            if self.expectations and self._expectations_checkpoint_location is None:
                warnings.warn(
                    f"Node '{self.name}' requires `expectations_checkpoint_location` specified unless Delta Live Tables is selected as an orchestrator and expectations are compatbile with DLT."
                )

        return self

    @model_validator(mode="after")
    def validate_view(self):
        if not self.is_view:
            return self

        # Validate Source
        if self.source:
            if not (
                isinstance(self.source, TableDataSource)
                or isinstance(self.source, PipelineNodeDataSource)
            ):
                raise ValueError(
                    "VIEW sink only supports Table or Pipeline Node with Table sink Data Source"
                )

            if self.source.as_stream:
                raise ValueError("VIEW sink does not support stream read.")

        # Validate Transformer
        view_defined = False
        if self.transformer:
            m = f"node '{self.name}': "

            if len(self.transformer.nodes) > 1:
                raise ValueError(
                    f"{m}VIEW sink only support transformers with a single SQL node"
                )

            if self.transformer.nodes[0].sql_expr is None:
                raise ValueError(
                    f"{m}VIEW sink only support transformers with a single SQL node"
                )

            view_defined = True

        # Validate Sinks
        for s in self.sinks:
            if s.table_type != "VIEW":
                raise ValueError(
                    f"{m}If one pipeline node sink is a VIEW, all of them must be a view."
                )

            if not view_defined:
                if s.view_definition is None:
                    raise ValueError(
                        "View definition must be provided at at the transformer or at the sink."
                    )

        # Validate Expectations
        if self.expectations:
            raise ValueError(f"{m}Expectations not supported for a view sink.")

        return self

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def child_attribute_names(self):
        return [
            "data_sources",
            "transformer",
            "sinks",
        ]

    # ----------------------------------------------------------------------- #
    # Orchestrator                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def is_orchestrator_dlt(self) -> bool:
        """If `True`, pipeline node is used in the context of a DLT pipeline"""
        pl = self.parent_pipeline
        if pl:
            return pl.is_orchestrator_dlt
        return False

    @property
    def is_dlt_run(self) -> bool:
        if not self.is_orchestrator_dlt:
            return False
        from laktory.dlt import is_debug

        return not is_debug()

    # ----------------------------------------------------------------------- #
    # Paths                                                                   #
    # ----------------------------------------------------------------------- #

    @property
    def _root_path(self) -> Path:
        if self.root_path:
            return Path(self.root_path)

        node = self.parent_pipeline
        if node and node._root_path:
            return node._root_path / self.name

        return Path(settings.laktory_root) / self.name

    # ----------------------------------------------------------------------- #
    # Outputs and Sinks                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def is_view(self) -> bool:
        if not self.sinks:
            return False
        is_view = False
        for s in self.sinks:
            if isinstance(s, TableDataSink) and s.table_type == "VIEW":
                is_view = True
                break
        return is_view

    @property
    def stage_df(self) -> AnyDataFrame:
        """
        Dataframe resulting from reading source and applying transformer, before data quality checks are applied.
        """
        return self._stage_df

    @property
    def output_df(self) -> AnyDataFrame:
        """
        Dataframe resulting from reading source, applying transformer and dropping rows not meeting data quality
        expectations.
        """
        return self._output_df

    @property
    def quarantine_df(self) -> AnyDataFrame:
        """
        DataFrame storing `stage_df` rows not meeting data quality expectations.
        """
        return self._quarantine_df

    @property
    def output_sinks(self) -> list[DataSinksUnion]:
        """List of sinks writing the output DataFrame"""
        sinks = []
        if self.sinks:
            for s in self.sinks:
                if not s.is_quarantine:
                    sinks += [s]
        return sinks

    @property
    def quarantine_sinks(self) -> list[DataSinksUnion]:
        """List of sinks writing the quarantine DataFrame"""
        sinks = []
        if self.sinks:
            for s in self.sinks:
                if s.is_quarantine:
                    sinks += [s]
        return sinks

    @property
    def all_sinks(self):
        """List of all sinks (output and quarantine)."""
        return self.output_sinks + self.quarantine_sinks

    @property
    def sinks_count(self) -> int:
        """Total number of sinks."""
        return len(self.all_sinks)

    @property
    def has_output_sinks(self) -> bool:
        """`True` if node has at least one output sink."""
        return len(self.output_sinks) > 0

    # @property
    # def has_quarantine_sinks(self) -> bool:
    #     """`True` if node has at least one quarantine sink."""
    #     return len(self.quarantine_sinks) > 0

    @property
    def has_sinks(self) -> bool:
        """`True` if node has at least one sink."""
        return self.sinks_count > 0

    @property
    def primary_sink(self) -> Union[DataSinksUnion, None]:
        """Primary output sink used as a source for downstream nodes."""
        if not self.has_output_sinks:
            return None

        s = None
        for s in self.output_sinks:
            if s.is_primary:
                break

        return s

    # ----------------------------------------------------------------------- #
    # Expectations                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def dlt_warning_expectations(self) -> dict[str, str]:
        expectations = {}
        for e in self.expectations:
            if e.is_dlt_compatible and e.action == "WARN":
                expectations[e.name] = e.expr.value
        return expectations

    @property
    def dlt_drop_expectations(self) -> dict[str, str]:
        expectations = {}
        for e in self.expectations:
            if e.is_dlt_compatible and e.action in ["DROP", "QUARANTINE"]:
                expectations[e.name] = e.expr.value
        return expectations

    @property
    def dlt_fail_expectations(self) -> dict[str, str]:
        expectations = {}
        for e in self.expectations:
            if e.is_dlt_compatible and e.action == "FAIL":
                expectations[e.name] = e.expr.value
        return expectations

    @property
    def _expectations_checkpoint_location(self) -> Path:
        if self.expectations_checkpoint_location:
            return Path(self.expectations_checkpoint_location)

        if self._root_path:
            return Path(self._root_path) / "checkpoints/expectations"

        return None

    @property
    def checks(self):
        return [e.check for e in self.expectations]

    # ----------------------------------------------------------------------- #
    # Transformations                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def layer_spark_chain(self):
        nodes = []

        if self.layer == "BRONZE":
            if self.add_layer_columns:
                nodes += [
                    SparkChainNode(
                        with_column={
                            "name": "_bronze_at",
                            "type": "timestamp",
                            "expr": "F.current_timestamp()",
                        },
                    ),
                ]

        elif self.layer == "SILVER":
            if self.timestamp_key:
                nodes += [
                    SparkChainNode(
                        with_column={
                            "name": "_tstamp",
                            "type": "timestamp",
                            "sql_expr": self.timestamp_key,
                        },
                    )
                ]

            if self.add_layer_columns:
                nodes += [
                    SparkChainNode(
                        with_column={
                            "name": "_silver_at",
                            "type": "timestamp",
                            "expr": "F.current_timestamp()",
                        },
                    )
                ]

        elif self.layer == "GOLD":
            if self.add_layer_columns:
                nodes += [
                    SparkChainNode(
                        with_column={
                            "name": "_gold_at",
                            "type": "timestamp",
                            "expr": "F.current_timestamp()",
                        },
                    )
                ]

        if self.drop_duplicates:
            subset = None
            if isinstance(self.drop_duplicates, list):
                subset = self.drop_duplicates
            elif self.primary_keys:
                subset = self.primary_keys

            nodes += [SparkChainNode(func_name="dropDuplicates", func_args=[subset])]

        if self.drop_source_columns and self.transformer:
            nodes += [
                SparkChainNode(
                    func_name="drop",
                    func_args=[
                        c
                        for c in self._source_columns
                        if c not in ["_bronze_at", "_silver_at", "_gold_at"]
                    ],
                )
            ]

        if len(nodes) == 0:
            return None

        return SparkChain(nodes=nodes)

    @property
    def layer_polars_chain(self):
        nodes = []

        if self.layer == "BRONZE":
            if self.add_layer_columns:
                nodes += [
                    PolarsChainNode(
                        with_column={
                            "name": "_bronze_at",
                            "type": "timestamp",
                            "expr": "pl.expr.laktory.current_timestamp()",
                        },
                    ),
                ]

        elif self.layer == "SILVER":
            if self.timestamp_key:
                nodes += [
                    PolarsChainNode(
                        with_column={
                            "name": "_tstamp",
                            "type": "timestamp",
                            "sql_expr": self.timestamp_key,
                        },
                    )
                ]

            if self.add_layer_columns:
                nodes += [
                    PolarsChainNode(
                        with_column={
                            "name": "_silver_at",
                            "type": "timestamp",
                            "expr": "pl.expr.laktory.current_timestamp()",
                        },
                    )
                ]

        elif self.layer == "GOLD":
            if self.add_layer_columns:
                nodes += [
                    PolarsChainNode(
                        with_column={
                            "name": "_gold_at",
                            "type": "timestamp",
                            "expr": "pl.expr.laktory.current_timestamp()",
                        },
                    )
                ]

        if self.drop_duplicates:
            subset = None
            if isinstance(self.drop_duplicates, list):
                subset = self.drop_duplicates
            elif self.primary_keys:
                subset = self.primary_keys

            nodes += [PolarsChainNode(func_name="unique", func_args=[subset])]

        if self.drop_source_columns and self.transformer:
            nodes += [
                PolarsChainNode(
                    func_name="drop",
                    func_args=[
                        c
                        for c in self._source_columns
                        if c not in ["_bronze_at", "_silver_at", "_gold_at"]
                    ],
                )
            ]

        if len(nodes) == 0:
            return None

        return PolarsChain(nodes=nodes)

    # ----------------------------------------------------------------------- #
    # Upstream Nodes                                                          #
    # ----------------------------------------------------------------------- #

    @property
    def upstream_node_names(self) -> list[str]:
        """Pipeline node names required to execute current node."""
        from laktory.models.datasources.pipelinenodedatasource import (
            PipelineNodeDataSource,
        )

        names = []

        if isinstance(self.source, PipelineNodeDataSource):
            names += [self.source.node_name]

        if self.transformer:
            names += self.transformer.upstream_node_names

        return names

    # ----------------------------------------------------------------------- #
    # Data Sources                                                            #
    # ----------------------------------------------------------------------- #

    @property
    def data_sources(self) -> list[BaseDataSource]:
        """Get all sources feeding the pipeline node"""
        sources = []

        if isinstance(self.source, BaseDataSource):
            sources += [self.source]

        if self.transformer:
            sources += self.transformer.data_sources

        if self.sinks:
            for s in self.sinks:
                if getattr(s, "view_definition", None):
                    sources += s.parsed_view_definition.data_sources

        return sources

    # ----------------------------------------------------------------------- #
    # Execution                                                               #
    # ----------------------------------------------------------------------- #

    def purge(self, spark=None):
        # TODO: Now that sink switch to overwrite when sink does not exists or when
        # a full refresh is requested, the purge method should not delete the data
        # by default, but only the checkpoints. Also consider truncating the table
        # instead of dropping it.

        logger.info(f"Purging pipeline node {self.name}")
        if self.has_sinks:
            for s in self.sinks:
                s.purge(spark=spark)
        if self._expectations_checkpoint_location:
            if os.path.exists(self._expectations_checkpoint_location):
                logger.info(
                    f"Deleting expectations checkpoint at {self._expectations_checkpoint_location}",
                )
                shutil.rmtree(self._expectations_checkpoint_location)

            if spark is None:
                return

            try:
                from pyspark.dbutils import DBUtils
            except ModuleNotFoundError:
                return

            dbutils = DBUtils(spark)

            _path = self._expectations_checkpoint_location.as_posix()
            try:
                dbutils.fs.ls(
                    _path
                )  # TODO: Figure out why this does not work with databricks connect
                logger.info(
                    f"Deleting checkpoint at dbfs {_path}",
                )
                dbutils.fs.rm(_path, True)

            except Exception as e:
                if "java.io.FileNotFoundException" in str(e):
                    pass
                elif "databricks.sdk.errors.platform.ResourceDoesNotExist" in str(
                    type(e)
                ):
                    pass
                elif "databricks.sdk.errors.platform.InvalidParameterValue" in str(
                    type(e)
                ):
                    # TODO: Figure out why this is happening. It seems that the databricks SDK
                    #       modify the path before sending to REST API.
                    logger.warn(f"dbutils could not delete checkpoint {_path}: {e}")
                else:
                    raise e

    def execute(
        self,
        apply_transformer: bool = True,
        spark: SparkSession = None,
        udfs: list[Callable] = None,
        write_sinks: bool = True,
        full_refresh: bool = False,
    ) -> AnyDataFrame:
        """
        Execute pipeline node by:

        - Reading the source
        - Applying the user defined (and layer-specific if applicable) transformations
        - Checking expectations
        - Writing the sinks

        Parameters
        ----------
        apply_transformer:
            Flag to apply transformer in the execution
        spark: SparkSession
            Spark session
        udfs:
            User-defined functions
        write_sinks:
            Flag to include writing sink in the execution
        full_refresh:
            If `True` dataframe will be completely re-processed by deleting
            existing data and checkpoint before processing.

        Returns
        -------
        :
            output Spark DataFrame
        """
        logger.info(f"Executing pipeline node {self.name}")

        # Parse DLT
        if self.is_orchestrator_dlt:
            logger.info("DLT orchestrator selected. Sinks writing will be skipped.")
            write_sinks = False
            full_refresh = False

        # Refresh
        if full_refresh:
            self.purge(spark)

        # Read Source
        self._stage_df = self.source.read(spark)

        # Save source
        self._source_columns = self._stage_df.columns

        # Apply transformer
        if apply_transformer:
            if self.is_view and self.transformer:
                self._view_definition = self.transformer.get_view_definition()

            if "spark" in str(type(self._stage_df)).lower():
                dftype = "spark"
            elif "polars" in str(type(self._stage_df)).lower():
                dftype = "polars"
            else:
                raise ValueError("DataFrame backend not supported")

            # Set Transformer
            transformer = self.transformer
            if transformer is None:
                if dftype == "spark":
                    transformer = SparkChain(nodes=[])
                elif dftype == "polars":
                    transformer = PolarsChain(nodes=[])
            else:
                transformer = transformer.model_copy()

            # Add layer-specific chain nodes
            if dftype == "spark" and self.layer_spark_chain:
                transformer.nodes += self.layer_spark_chain.nodes
            elif dftype == "polars" and self.layer_polars_chain:
                transformer.nodes += self.layer_polars_chain.nodes

            if transformer.nodes:
                self._stage_df = transformer.execute(self._stage_df, udfs=udfs)

        # Check expectations
        self.check_expectations()

        # Output and Quarantine to Sinks
        if write_sinks:
            for s in self.output_sinks:
                if self.is_view:
                    s.write(view_definition=self._view_definition, spark=spark)
                    self._output_df = s.as_source().read(spark=spark)
                else:
                    s.write(self._output_df, full_refresh=full_refresh)
            if self._quarantine_df is not None:
                for s in self.quarantine_sinks:
                    s.write(self._quarantine_df, full_refresh=full_refresh)

        return self._output_df

    def check_expectations(self):
        """
        Check expectations, raise errors, warnings where required and build
        filtered and quarantine DataFrames.

        Some actions have to be disabled when selected orchestrator is
        Databricks DLT:

        * Raising error on Failure when expectation is supported by DLT
        * Dropping rows when expectation is supported by DLT
        """

        # Data Quality Checks
        is_streaming = getattr(self._stage_df, "isStreaming", False)
        qfilter = None  # Quarantine filter
        kfilter = None  # Keep filter
        if not self.expectations:
            self._output_df = self._stage_df
            self._quarantine_df = None
            return

        logger.info("Checking Data Quality Expectations")

        def _batch_check(df, node):
            for e in node.expectations:
                is_dlt_managed = node.is_dlt_run and e.is_dlt_compatible

                # Run Check
                if not is_dlt_managed:
                    e.run_check(
                        df,
                        raise_or_warn=True,
                        node=node,
                    )

        def _stream_check(batch_df, batch_id, node):
            _batch_check(
                batch_df,
                node,
            )

        # Warn or Fail
        if is_streaming and self.is_dlt_run:
            # TODO: Enable when DLT supports foreachBatch (in case some expectations are not supported by DLT)
            pass

        elif is_streaming:
            if self._expectations_checkpoint_location is None:
                raise ValueError(
                    f"Expectations Checkpoint not specified for node '{self.name}'"
                )
            query = (
                self._stage_df.writeStream.foreachBatch(
                    lambda batch_df, batch_id: _stream_check(batch_df, batch_id, self)
                )
                .trigger(availableNow=True)
                .options(
                    checkpointLocation=self._expectations_checkpoint_location,
                )
                .start()
            )
            query.awaitTermination()

        else:
            _batch_check(
                self._stage_df,
                self,
            )

        # Build Filters
        for e in self.expectations:
            is_dlt_managed = self.is_dlt_run and e.is_dlt_compatible

            # Update Keep Filter
            if not is_dlt_managed:
                _filter = e.keep_filter
                if _filter is not None:
                    if kfilter is None:
                        kfilter = _filter
                    else:
                        kfilter = kfilter & _filter

            # Update Quarantine Filter
            _filter = e.quarantine_filter
            if _filter is not None:
                if qfilter is None:
                    qfilter = _filter
                else:
                    qfilter = qfilter & _filter

        if qfilter is not None:
            logger.info("Building quarantine DataFrame")
            self._quarantine_df = self._stage_df.filter(qfilter)
        else:
            self._quarantine_df = self._stage_df.filter("False")

        if kfilter is not None:
            logger.info("Dropping invalid rows")
            self._output_df = self._stage_df.filter(kfilter)
        else:
            self._output_df = self._stage_df



from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class ServicePrincipalRole(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Service Principal role

    Attributes
    ----------
    role:
        This is the id of the role or instance profile resource.
    service_principal_id:
        This is the id of the service principal resource.
    """

    role: str = None
    service_principal_id: str = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return f"{self.role}-{self.service_principal_id}"

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:ServicePrincipalRole"

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_service_principal_role"



import pyspark.sql.types as T

DATATYPES_MAP = {
    "binary": T.BinaryType(),
    "byte": T.ByteType(),
    "int8": T.ByteType(),
    "tinyint": T.ByteType(),
    "short": T.ShortType(),
    "int16": T.ShortType(),
    "smallint": T.ShortType(),
    "int": T.IntegerType(),
    "int32": T.IntegerType(),
    "long": T.LongType(),
    "int64": T.LongType(),
    "bigint": T.LongType(),
    "float": T.FloatType(),
    "float32": T.FloatType(),
    "double": T.DoubleType(),
    "float64": T.DoubleType(),
    "boolean": T.BooleanType(),
    "string": T.StringType(),
    "utf8": T.StringType(),
    "date": T.DateType(),
    "timestamp": T.TimestampType(),
    "datetime": T.TimestampType(),
    # "struct<>": T.StructType(),
    # "struct": T.StructType(),
    # "array": T.ArrayType(),
    # "list": T.ArrayType(),
}



from functools import wraps

import polars as pl

from laktory.polars.dataframe.groupby_and_agg import groupby_and_agg
from laktory.polars.dataframe.has_column import has_column
from laktory.polars.dataframe.schema_flat import schema_flat
from laktory.polars.dataframe.signature import signature
from laktory.polars.dataframe.smart_join import smart_join
from laktory.polars.dataframe.union import union
from laktory.polars.dataframe.window_filter import window_filter


@pl.api.register_dataframe_namespace("laktory")
class LaktoryDataFrame:  # noqa: F811
    def __init__(self, df: pl.DataFrame):
        self._df = df

    @wraps(groupby_and_agg)
    def groupby_and_agg(self, *args, **kwargs):
        return groupby_and_agg(self._df, *args, **kwargs)

    @wraps(has_column)
    def has_column(self, *args, **kwargs):
        return has_column(self._df, *args, **kwargs)

    @wraps(signature)
    def signature(self, *args, **kwargs):
        return signature(self._df, *args, **kwargs)

    @wraps(schema_flat)
    def schema_flat(self, *args, **kwargs):
        return schema_flat(self._df, *args, **kwargs)

    @wraps(smart_join)
    def smart_join(self, *args, **kwargs):
        return smart_join(self._df, *args, **kwargs)

    @wraps(union)
    def union(self, *args, **kwargs):
        return union(self._df, *args, **kwargs)

    @wraps(window_filter)
    def window_filter(self, *args, **kwargs):
        return window_filter(self._df, *args, **kwargs)


@pl.api.register_lazyframe_namespace("laktory")
class LaktoryDataFrame:  # noqa: F811
    def __init__(self, df: pl.LazyFrame):
        self._df = df

    @wraps(groupby_and_agg)
    def groupby_and_agg(self, *args, **kwargs):
        return groupby_and_agg(self._df, *args, **kwargs)

    @wraps(has_column)
    def has_column(self, *args, **kwargs):
        return has_column(self._df, *args, **kwargs)

    @wraps(signature)
    def signature(self, *args, **kwargs):
        return signature(self._df, *args, **kwargs)

    @wraps(schema_flat)
    def schema_flat(self, *args, **kwargs):
        return schema_flat(self._df, *args, **kwargs)

    @wraps(smart_join)
    def smart_join(self, *args, **kwargs):
        return smart_join(self._df, *args, **kwargs)

    @wraps(union)
    def union(self, *args, **kwargs):
        return union(self._df, *args, **kwargs)

    @wraps(window_filter)
    def window_filter(self, *args, **kwargs):
        return window_filter(self._df, *args, **kwargs)



import polars as pl


def signature(df: pl.DataFrame) -> str:
    """
    Returns DataFrame signature.

    Parameters
    ----------
    df:
        Input DataFrame

    Returns
    -------
    :
        Result

    ```py
    import polars as pl

    import laktory  # noqa: F401

    df = pl.DataFrame({"x": [1, 2, 3], "label": ["a", "b", "c"]})
    print(df.laktory.signature())
    # > DataFrame[x: Int64, label: String]
    ```

    """

    s = ", ".join([f"{k}: {v}" for k, v in df.schema.items()])
    s = f"DataFrame[{s}]"
    return s



from typing import Union

from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.mwsnccbinding import MwsNccBinding
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class MwsNetworkConnectivityConfigEgressConfigDefaultRulesAwsStableIpRule(BaseModel):
    """
    Attributes
    ----------
    cidr_blocks:
        todo
    """

    cidr_blocks: list[str] = None


class MwsNetworkConnectivityConfigEgressConfigDefaultRulesAzureServiceEndpointRule(
    BaseModel
):
    """
    Attributes
    ----------
    subnets:
        todo
    target_region:
        todo
    target_services:
        todo
    """

    subnets: list[str] = None
    target_region: str = None
    target_services: list[str] = None


class MwsNetworkConnectivityConfigEgressConfigDefaultRules(BaseModel):
    """
    Attributes
    ----------
    aws_stable_ip_rule:
        todo
    azure_service_endpoint_rule:
        This provides a list of subnets. These subnets need to be allowed in your Azure resources in order for
        Databricks to access.
    """

    aws_stable_ip_rule: MwsNetworkConnectivityConfigEgressConfigDefaultRulesAwsStableIpRule = None
    azure_service_endpoint_rule: MwsNetworkConnectivityConfigEgressConfigDefaultRulesAzureServiceEndpointRule = None


class MwsNetworkConnectivityConfigEgressConfigTargetRulesAzurePrivateEndpointRule(
    BaseModel
):
    """
    Attributes
    ----------
    connection_state:
        todo
    creation_time:
        todo
    deactivated:
        todo
    deactivated_at:
        todo
    endpoint_name:
        todo
    group_id:
        todo
    network_connectivity_config_id:
        Canonical unique identifier of Network Connectivity Config in Databricks Account
    resource_id:
        todo
    rule_id:
        todo
    updated_time:
        todo
    """

    connection_state: str = None
    creation_time: int = None
    deactivated: bool = None
    deactivated_at: int = None
    endpoint_name: str = None
    group_id: str = None
    network_connectivity_config_id: str = None
    resource_id: str = None
    rule_id: str = None
    updated_time: int = None


class MwsNetworkConnectivityConfigEgressConfigTargetRules(BaseModel):
    """
    Attributes
    ----------
    azure_private_endpoint_rules:
        todo
    """

    azure_private_endpoint_rules: list[
        MwsNetworkConnectivityConfigEgressConfigTargetRulesAzurePrivateEndpointRule
    ] = None


class MwsNetworkConnectivityConfigEgressConfig(BaseModel):
    """
    Databricks Metastore Data Access AWS IAM Role

    Attributes
    ----------
    default_rules:
        todo
    target_rules:
        todo
    """

    default_rules: MwsNetworkConnectivityConfigEgressConfigDefaultRules = None
    target_rules: MwsNetworkConnectivityConfigEgressConfigTargetRules = None


class MwsNetworkConnectivityConfig(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Mws Network Connectivity Config

    Attributes
    ----------
    account_id:
        todo
    creation_time:
        todo
    egress_config:
        Egress Config configration
    name:
        Name of Network Connectivity Config in Databricks Account. Change forces creation of a new resource.
    network_connectivity_config_id:
        Canonical unique identifier of Network Connectivity Config in Databricks Account
    region:
        Region of the Network Connectivity Config. NCCs can only be referenced by your workspaces in the same region.
        Change forces creation of a new resource.
    updated_time:
        todo

    Examples
    --------
    ```py
    ```
    """

    account_id: str = None
    creation_time: int = None
    egress_config: MwsNetworkConnectivityConfigEgressConfig = None
    name: str
    network_connectivity_config_id: str = None
    region: str
    updated_time: int = None
    workspace_bindings: list[MwsNccBinding] = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - workspace bindings
        """
        resources = []

        if self.workspace_bindings:
            for b in self.workspace_bindings:
                b.network_connectivity_config_id = f"${{resources.{self.resource_name}.network_connectivity_config_id}}"
                resources += [b]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:MwsNetworkConnectivityConfig"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return [
            "workspace_bindings",
        ]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_mws_network_connectivity_config"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Literal
from typing import Union

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.pipeline.pipelinechild import PipelineChild
from laktory.models.transformers.basechainnode import BaseChainNode
from laktory.typing import AnyDataFrame

if TYPE_CHECKING:
    pass

logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class BaseChain(BaseModel, PipelineChild):
    dataframe_backend: Literal["SPARK", "POLARS"] = None
    nodes: list[Union[BaseChainNode, "BaseChain"]]
    _columns: list[list[str]] = []

    @property
    def columns(self):
        return self._columns

    @property
    def upstream_node_names(self) -> list[str]:
        """Pipeline node names required to apply transformer"""
        names = []
        for node in self.nodes:
            names += node.upstream_node_names
        return names

    @property
    def data_sources(self):
        """Get all sources feeding the Transformer"""
        sources = []
        for node in self.nodes:
            sources += node.data_sources
        return sources

    # ----------------------------------------------------------------------- #
    # Children                                                                #
    # ----------------------------------------------------------------------- #

    @property
    def child_attribute_names(self):
        return ["nodes"]

    # ----------------------------------------------------------------------- #
    # Execution                                                               #
    # ----------------------------------------------------------------------- #

    def execute(self, df, udfs=None) -> AnyDataFrame:
        logger.info(f"Executing {self.df_backend} chain")

        for inode, node in enumerate(self.nodes):
            self._columns += [df.columns]

            tnode = type(node)
            logger.info(
                f"Executing {self.df_backend} chain node {inode} ({tnode.__name__})."
            )
            df = node.execute(df, udfs=udfs)

        return df

    def get_view_definition(self):
        logger.info("Creating view definition")
        return self.nodes[0].get_view_definition()


BaseModel.model_rebuild()



# COMMAND ----------

dbutils.widgets.text("pipeline_name", "pl-stocks-job")
dbutils.widgets.text("node_name", "")
dbutils.widgets.text("full_refresh", "False")
dbutils.widgets.text("install_dependencies", "True")

# COMMAND ----------
install_dependencies = dbutils.widgets.get("install_dependencies").lower() == "true"
pl_name = dbutils.widgets.get("pipeline_name")

if install_dependencies:
    notebook_path = (
        dbutils.notebook.entry_point.getDbutils()
        .notebook()
        .getContext()
        .notebookPath()
        .get()
    )
    laktory_root = "/Workspace" + notebook_path.split("/jobs/")[0]
    filepath = f"{laktory_root}/pipelines/{pl_name}/requirements.txt"
    # MAGIC     %pip install -r $filepath
    # MAGIC     %restart_python

# COMMAND ----------
import importlib
import os
import sys

import pyspark.sql.functionas as F  # noqa: F401

from laktory import get_logger
from laktory import models

logger = get_logger(__name__)
notebook_path = (
    dbutils.notebook.entry_point.getDbutils()
    .notebook()
    .getContext()
    .notebookPath()
    .get()
)

# --------------------------------------------------------------------------- #
# Read Pipeline                                                               #
# --------------------------------------------------------------------------- #

laktory_root = "/Workspace" + notebook_path.split("/jobs/")[0]
pl_name = dbutils.widgets.get("pipeline_name")
node_name = dbutils.widgets.get("node_name")
full_refresh = dbutils.widgets.get("full_refresh").lower() == "true"
filepath = f"{laktory_root}/pipelines/{pl_name}/config.json"
with open(filepath, "r") as fp:
    pl = models.Pipeline.model_validate_json(fp.read())

# Import User Defined Functions
sys.path.append(f"{laktory_root}/pipelines/")
udfs = []
for udf in pl.udfs:
    if udf.module_path:
        sys.path.append(os.path.abspath(udf.module_path))
    module = importlib.import_module(udf.module_name)
    module = importlib.reload(module)
    globals()[udf.module_name] = module
    udfs += [getattr(module, udf.function_name)]

# --------------------------------------------------------------------------- #
# Execution                                                                   #
# --------------------------------------------------------------------------- #

if node_name:
    pl.nodes_dict[node_name].execute(spark=spark, udfs=udfs, full_refresh=full_refresh)
else:
    pl.execute(spark=spark, udfs=udfs, full_refresh=full_refresh)



from laktory.models.resources.providers.baseprovider import BaseProvider
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class AzureProvider(BaseProvider, PulumiResource, TerraformResource):
    """
    Azure Provider

    Attributes
    ----------
    auxiliary_tenant_ids:
        #TODO
    client_certificate:
        Base64 encoded PKCS#12 certificate bundle to use when authenticating
        as a Service Principal using a Client Certificate
    client_certificate_password:
        The password associated with the Client Certificate. For use when
        authenticating as a Service Principal using a Client Certificate
    client_certificate_path:
        The path to the Client Certificate associated with the Service
        Principal for use when authenticating as a Service Principal using
        a Client Certificate.
    client_id:
        The Client ID which should be used.
    client_id_file_path:
        The path to a file containing the Client ID which should be used.
    client_secret:
        The Client Secret which should be used. For use When authenticating as
        a Service Principal using a Client Secret.
    client_secret_file_path:
        The path to a file containing the Client Secret which should be used.
        For use When authenticating as a Service Principal using a Client
        Secret.
    disable_correlation_request_id:
        This will disable the x-ms-correlation-request-id header.
    disable_terraform_partner_id:
        This will disable the Terraform Partner ID which is used if a custom
        partner_id isn't specified.
    environment:
        The Cloud Environment which should be used. Possible values are public,
        usgovernment, and china. Defaults to public. It can also be sourced
        from the following environment variables: AZURE_ENVIRONMENT,
        ARM_ENVIRONMENT
    metadata_host:
        The Hostname which should be used for the Azure Metadata Service. It
        can also be sourced from the following environment variable:
        ARM_METADATA_HOSTNAME
    msi_endpoint:
        The path to a custom endpoint for Managed Service Identity - in most
        circumstances this should be detected automatically.
    oidc_request_token:
        The bearer token for the request to the OIDC provider. For use
        when authenticating as a Service Principal using OpenID Connect.
    oidc_request_url:
        The URL for the OIDC provider from which to request an ID token. For
        use when authenticating as a Service Principal using OpenID Connect.
    oidc_token:
        The OIDC ID token for use when authenticating as a Service Principal
        using OpenID Connect.
    oidc_token_file_path:
        The path to a file containing an OIDC ID token for use when
        authenticating as a Service Principal using OpenID Connect.
    partner_id:
        A GUID/UUID that is registered with Microsoft to facilitate partner
        resource usage attribution.
    skip_provider_registration:
        Should the AzureRM Provider skip registering all of the Resource
        Providers that it supports, if they're not already registered? It can
        also be sourced from the following environment variable:
        ARM_SKIP_PROVIDER_REGISTRATION
    storage_use_azuread:
        Should the AzureRM Provider use AzureAD to access the Storage Data
        Plane API's? It can also be sourced from the following environment
        variable: ARM_STORAGE_USE_AZUREAD
    subscription_id:
        The Subscription ID which should be used. It can also be sourced from
        the following environment variable: ARM_SUBSCRIPTION_ID
    tenant_id:
        The Tenant ID which should be used.
    use_aks_workload_identity:
        Allow Azure AKS Workload Identity to be used for Authentication.
    use_cli:
        Allow Azure CLI to be used for Authentication.
    use_msi:
        Allow Managed Service Identity to be used for Authentication.
    use_oidc:
        Allow OpenID Connect to be used for authentication


    Examples
    --------
    ```py
    from laktory import models

    p = models.AzureProvider(
        client_id="${vars.AZURE_CLIENT_ID}",
        client_secret="${vars.AZURE_CLIENT_SECRET}",
    )
    ```
    """

    auxiliary_tenant_ids: list[str] = None
    client_certificate: str = None
    client_certificate_password: str = None
    client_certificate_path: str = None
    client_id: str = None
    client_id_file_path: str = None
    client_secret: str = None
    client_secret_file_path: str = None
    disable_correlation_request_id: bool = None
    disable_terraform_partner_id: bool = None
    environment: str = None
    # features: ProviderFeaturesArgs = None
    metadata_host: str = None
    msi_endpoint: str = None
    oidc_request_token: str = None
    oidc_request_url: str = None
    oidc_token: str = None
    oidc_token_file_path: str = None
    partner_id: str = None
    skip_provider_registration: bool = None
    storage_use_azuread: bool = None
    subscription_id: str = None
    tenant_id: str = None
    use_aks_workload_identity: bool = None
    use_cli: bool = None
    use_msi: bool = None
    use_oidc: bool = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    # @property
    # def resource_key(self) -> str:
    #     return self.display_name

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "pulumi:providers:azure"



from typing import Literal
from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class WarehouseCustomTag(BaseModel):
    """
    Warehouse Custom Tag specifications

    Attributes
    ----------
    key:
        Tag key
    value:
        Tag value
    """

    key: str
    value: str


class WarehouseTags(BaseModel):
    """
    Warehouse Tags specifications

    Attributes
    ----------
    custom_tags:
        Tags specifications
    """

    custom_tags: list[WarehouseCustomTag] = []

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "custom_tags": "custom_tags",
        }


class WarehouseLookup(ResourceLookup):
    """
    Attributes
    ----------
    id:
        The ID of the SQL warehouse.
    """

    id: str = Field(serialization_alias="id")


class Warehouse(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Warehouse

    Attributes
    ----------
    access_controls:
        Warehouse access controls
    auto_stop_mins:
        Time in minutes until an idle SQL warehouse terminates all clusters and stops.
    channel_name:
        Channel specifications
    cluster_size:
        The size of the clusters allocated to the endpoint
    enable_photon:
        If `True`, photon is enabled
    enable_serverless_compute:
        If `True`, warehouse is serverless.
        See details [here](https://www.pulumi.com/registry/packages/databricks/api-docs/sqlendpoint/)
    instance_profile_arn:
        TODO
    jdbc_url:
        JDBC connection string.
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    max_num_clusters:
        Maximum number of clusters available when a SQL warehouse is running.
    min_num_clusters:
        Minimum number of clusters available when a SQL warehouse is running.
    name:
        Warehouse name
    num_clusters:
        Fixed number of clusters when autoscaling is not enabled.
    spot_instance_policy:
        The spot policy to use for allocating instances to clusters.
    tags:
        Databricks tags all endpoint resources with these tags.
    warehouse_type:
        SQL warehouse type.

    Examples
    --------
    ```py
    from laktory import models

    warehouse = models.resources.databricks.Warehouse(
        name="default",
        cluster_size="2X-Small",
        auto_stop_mins=30,
        channel_name="CHANNEL_NAME_PREVIEW",
        enable_photon=True,
        enable_serverless_compute=True,
        access_controls=[
            {"group_name": "account users", "permission_level": "CAN_USE"}
        ],
    )
    ```
    """

    cluster_size: Literal[
        "2X-Small",
        "X-Small",
        "Small",
        "Medium",
        "Large",
        "X-Large",
        "2X-Large",
        "3X-Large",
        "4X-Large",
    ]
    access_controls: list[AccessControl] = []
    auto_stop_mins: int = None
    channel_name: Union[
        Literal["CHANNEL_NAME_CURRENT", "CHANNEL_NAME_PREVIEW"], str
    ] = None
    # data_source_id
    enable_photon: bool = None
    enable_serverless_compute: bool = None
    instance_profile_arn: str = None
    jdbc_url: str = None
    lookup_existing: WarehouseLookup = Field(None, exclude=True)
    max_num_clusters: int = None
    min_num_clusters: int = None
    name: str
    num_clusters: int = None
    # odbc_params
    spot_instance_policy: Literal["COST_OPTIMIZED", "RELIABILITY_OPTIMIZED"] = None
    # state
    tags: WarehouseTags = None
    warehouse_type: Literal["CLASSIC", "PRO"] = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - warehouse permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    sql_endpoint_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:SqlEndpoint"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls"]

    @property
    def pulumi_properties(self):
        d = super().pulumi_properties
        if self._camel_serialization:
            d["channel"] = {"name": d.pop("channelName", None)}
        else:
            d["channel"] = {"name": d.pop("channel_name", None)}
        return d

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_sql_endpoint"

    @property
    def terraform_resource_lookup_type(self) -> str:
        return "databricks_sql_warehouse"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes

    @property
    def terraform_properties(self) -> dict:
        d = super().terraform_properties
        d["channel"] = {"name": d.pop("channel_name", None)}
        return d



from typing import Literal

from laktory.models.basemodel import BaseModel
from laktory.models.resources.pulumiresource import PulumiResource


class StorageBlob(BaseModel, PulumiResource):
    """
    Databricks group

    Attributes
    ----------
    access_tier:
        The access tier of the storage blob.
    account_name:
        Specifies the storage account in which to create the storage container.
    blob_name:
        The name of the storage blob. Must be unique within the storage container the blob is located.
    container_name:
        The name of the storage container in which this blob should be created.
    content_md5:
        The MD5 sum of the blob contents. Cannot be defined if blob type is Append.
    content_type:
        The content type of the storage blob. Defaults to application/octet-stream.
    metadata:
        A map of custom blob metadata.
    resource_group_name:
        The name of the resource group within the user's subscription.
    type:
        The type of the storage blob to be created. Defaults to 'Block'.
    """

    access_tier: Literal["HOT", "COOL", "ARCHIVE"] = None
    account_name: str = None
    blob_name: str = None
    container_name: str = None
    content_md5: str = None
    content_type: str = None
    metadata: dict[str, str] = None
    resource_group_name: str = None
    # source: Union[pulumi.Asset, pulumi.Archive] = None
    type: Literal["BLOCK", "APPEND"] = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """Blob name"""
        return self.blob_name

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "azure-native:storage:Blob"

    # # ----------------------------------------------------------------------- #
    # # Terraform Properties                                                    #
    # # ----------------------------------------------------------------------- #
    #
    # @property
    # def terraform_resource_type(self) -> str:
    #     return "databricks_group"
    #
    # @property
    # def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
    #     return self.pulumi_excludes



from typing import Union

from pyspark.sql.dataframe import DataFrame


def show_string(
    df: DataFrame,
    n: int = 20,
    truncate: Union[bool, int] = True,
    vertical: bool = False,
) -> str:
    """
    Returns the string generated by `df.show()`

    Parameters
    ----------
    n:
        Number of rows to show.
    truncate:
        If set to `True`, truncate strings longer than 20 chars by default. If
        set to a number greater than one, truncates long strings to length
        truncate and align cells right.
    vertical:
        If set to True, print output rows vertically (one line per column
        value).

    Returns
    -------
    :
        Show string

    Examples
    --------

    ```py
    import laktory  # noqa: F401

    df = spark.createDataFrame([[7, "a"], [8, "b"], [9, "c"]], ["x", "y"])
    print(df.laktory.show_string())
    '''
    +---+---+
    |  x|  y|
    +---+---+
    |  7|  a|
    |  8|  b|
    |  9|  c|
    +---+---+
    '''
    ```
    """
    if truncate is True:
        truncate = 20
    return df._jdf.showString(n, int(truncate), vertical)



import os


class MonkeyPatch:
    def __init__(self):
        self.env0 = {}

    def setenv(self, key, value):
        self.env0[key] = os.getenv(key, None)
        os.environ[key] = value

    def cleanup(self):
        for k, v in self.env0.items():
            if v is None:
                del os.environ[k]
            else:
                os.environ[k] = v



from __future__ import annotations

import time
from typing import TYPE_CHECKING
from typing import Literal

from laktory._logger import get_logger
from laktory.datetime import unix_timestamp
from laktory.dispatcher.dispatcherrunner import DispatcherRunner

if TYPE_CHECKING:
    from databricks.sdk.service.pipelines import GetUpdateResponse
    from databricks.sdk.service.pipelines import StartUpdateResponse

logger = get_logger(__name__)


class DLTPipelineRunner(DispatcherRunner):
    """
    DLT Pipeline runner.
    """

    _update_start: StartUpdateResponse = None
    _update: GetUpdateResponse = None

    def get_id(self) -> str:
        """Get deployed pipeline id"""
        logger.info(f"Getting id for pipeline {self.name}")
        for _pl in self.wc.pipelines.list_pipelines(
            filter=f"name LIKE '{self.name}'", max_results=1
        ):
            if self.name == _pl.name:
                self.id = _pl.pipeline_id
                return self.id

    def run(
        self,
        wait: bool = True,
        timeout: int = 20 * 60,
        full_refresh: bool = False,
        raise_exception: bool = False,
        current_run_action: Literal["WAIT", "CANCEL", "FAIL"] = "WAIT",
    ):
        """
        Run remote pipeline and monitor failures.

        Parameters
        ----------
        wait:
            If `True`, runs synchronously and wait for completion.
        timeout:
            Maximum time allowed for the pipeline to run before an exception is
            raised.
        full_refresh:
            If `True`, tables are fully refreshed (re-built). Otherwise, only
            increments are processed.
        raise_exception:
            If `True`, exceptions are raised if the pipeline fails.
        current_run_action:
            Action to take with respect to current run (if any). Possible
            options are:
                - WAIT: wait for the current run to complete
                - CANCEL: cancel the current run
                - FAIL: raise an exception

        Returns
        -------
        output:
            None
        """
        from databricks.sdk.core import DatabricksError
        from databricks.sdk.service.pipelines import EventLevel
        from databricks.sdk.service.pipelines import UpdateInfoState

        event_ids = []

        # Start update
        t0 = time.time()
        try:
            self._update_start = self.wc.pipelines.start_update(
                pipeline_id=self.id,
                full_refresh=full_refresh,
                # validate_only=False,
            )
            logger.info(f"Pipeline {self.name} update started...")

        except DatabricksError as e:
            # Exception is raised when update is currently in progress
            if current_run_action.upper() == "FAIL":
                logger.info(f"Pipeline {self.name} already running...")
                raise e

            elif current_run_action.upper() == "WAIT":
                logger.info(
                    f"Pipeline {self.name} waiting for current update to be completed..."
                )
                self.wc.pipelines.wait_get_pipeline_idle(pipeline_id=self.id)

            elif current_run_action.upper() == "CANCEL":
                logger.info(f"Pipeline {self.name} cancelling current update...")
                self.wc.pipelines.stop_and_wait(pipeline_id=self.id)

            logger.info(f"Pipeline {self.name} update started...")
            self._update_start = self.wc.pipelines.start_update(
                pipeline_id=self.id,
                full_refresh=full_refresh,
                # validate_only=False,
            )

        pstate = None
        self.get_update()
        logger.info(f"Pipeline {self.name} run URL: {self.update_url}")
        if wait:
            while (
                time.time() - t0 < timeout
                or self.update_state == UpdateInfoState.COMPLETED
            ):
                self.get_update()

                if self.update_state != pstate:
                    logger.info(
                        f"Pipeline {self.name} state: {self.update_state.value}"
                    )
                    pstate = self.update_state

                # filter don't seem to work
                for event in self.wc.pipelines.list_pipeline_events(
                    pipeline_id=self.id,
                    max_results=100,
                    # filter=f"(level in ('ERROR', 'WARN')) AND (timestamp > '{utc_datetime(t0).isoformat()}Z')"
                    # filter=f"level in ('ERROR', 'WARN')",
                    # filter=f"timestamp > '{utc_datetime(t0).isoformat()}Z'"
                    page_token=None,
                ):
                    if (
                        event.id in event_ids
                        or unix_timestamp(event.timestamp) < t0
                        or event.origin.update_id != self.update_id
                    ):
                        continue

                    if event.level == EventLevel.WARN:
                        logger.warn(event.message)
                    elif event.level == EventLevel.ERROR:
                        logger.error(event.message)
                    event_ids += [event.id]

                if self.update_state in [
                    UpdateInfoState.CANCELED,
                    UpdateInfoState.COMPLETED,
                    UpdateInfoState.FAILED,
                ]:
                    break

                time.sleep(1.0)

            logger.info(
                f"Pipeline {self.name} update terminated after {time.time() - t0: 5.2f} sec with {self.update_state}"
            )
            if raise_exception and self.update_state != UpdateInfoState.COMPLETED:
                raise Exception(
                    f"Pipeline {self.name} update not completed ({self.update_state})"
                )

    def get_update(self):
        self._update = self.wc.pipelines.get_update(
            pipeline_id=self.id, update_id=self.update_id
        )
        return self._update

    @property
    def update_state(self):
        return self._update.update.state

    @property
    def update_id(self) -> str:
        return self._update_start.update_id

    @property
    def update_url(self) -> str:
        return f"{self.wc.config.host}/pipelines/{self.id}/updates/{self.update_id}"



# import pyspark.sql.functions as F #noqa
from planck import units
from pyspark.sql.column import Column

from laktory.spark.functions._common import COLUMN_OR_NAME
from laktory.spark.functions._common import _col

__all__ = [
    "convert_units",
]


# --------------------------------------------------------------------------- #
# Polynomials                                                                 #
# --------------------------------------------------------------------------- #


def convert_units(
    x: COLUMN_OR_NAME,
    input_unit: str,
    output_unit: str,
) -> Column:
    """
    Units conversion

    Parameters
    ----------
    x:
        Input column
    input_unit:
        Input units
    output_unit:
        Output units

    Returns
    -------
    :
        Output column

    Examples
    --------
    ```py
    import pyspark.sql.functions as F

    import laktory  # noqa: F401

    df = spark.createDataFrame([[1.0]], ["x"])
    df = df.withColumn(
        "y", F.laktory.convert_units("x", input_unit="m", output_unit="ft")
    )
    print(df.laktory.show_string())
    '''
    +---+-----------------+
    |  x|                y|
    +---+-----------------+
    |1.0|3.280839895013124|
    +---+-----------------+
    '''
    ```

    References
    ----------
    The units conversion function use [planck](https://www.okube.ai/planck/) convert as a backend.
    """
    return units.convert(_col(x), input_unit, output_unit)


if __name__ == "__main__":
    import pyspark.sql.functions as F

    import laktory  # noqa: F401
    from laktory._testing import sparkf

    spark = sparkf.spark

    df = spark.createDataFrame([[1.0]], ["x"])
    df = df.withColumn(
        "y", F.laktory.convert_units("x", input_unit="m", output_unit="ft")
    )
    print(df.laktory.show_string())



from typing import Union

import polars as pl

__all__ = [
    "compare",
]


# --------------------------------------------------------------------------- #
# compare                                                                     #
# --------------------------------------------------------------------------- #


def compare(
    x: pl.Expr,
    y: Union[pl.Expr, float, str, bool] = 0,
    where: pl.Expr = None,
    operator: str = "==",
    default: Union[pl.Expr, float, str, bool] = None,
) -> pl.Expr:
    """
    Compare a column `x` and a value or another column `y` using
    `operator`. Comparison can be limited to `where` and assigned
    `default` elsewhere.

    output = `x` `operator` `y`

    Parameters
    ---------
    x :
        Base column to compare
    y :
        Column to compare to
    where:
        Where to apply the comparison
    operator: str
        Operator for comparison
    default:
        Default value to be applied when `where` is `False`

    Returns
    -------
    :
        Comparison result

    Examples
    --------
    ```py
    import polars as pl

    import laktory  # noqa: F401

    df = pl.DataFrame({"x": [0.45, 0.55]})
    df = df.with_columns(
        y=pl.expr.laktory.compare(
            pl.col("x"),
            pl.lit(0.5),
            operator=">",
        )
    )
    print(df.glimpse(return_as_string=True))
    '''
    Rows: 2
    Columns: 2
    $ x  <f64> 0.45, 0.55
    $ y <bool> False, True
    '''
    ```
    """
    if operator == "==":
        c = x == y
    elif operator == "!=":
        c = x != y
    elif operator == "<":
        c = x < y
    elif operator == "<=":
        c = x <= y
    elif operator == ">":
        c = x > y
    elif operator == ">=":
        c = x >= y
    else:
        raise ValueError(f"Operator '{operator}' is not supported.")

    if where is not None:
        if default is None:
            default = pl.lit(None)

        c = pl.when(where).then(c).otherwise(default)

    return c



import json
import os
import re
from collections import defaultdict
from typing import Any
from typing import Union

from pydantic import model_validator

from laktory._logger import get_logger
from laktory._parsers import _resolve_values
from laktory._settings import settings
from laktory._useragent import set_databricks_sdk_upstream
from laktory.constants import CACHE_ROOT
from laktory.models.basemodel import BaseModel
from laktory.models.resources.providers.baseprovider import BaseProvider

logger = get_logger(__name__)


class ConfigValue(BaseModel):
    type: str = "String"
    description: str = None
    default: Any = None


class TerraformRequiredProvider(BaseModel):
    source: str
    version: Union[str, None] = None


class TerraformConfig(BaseModel):
    required_providers: dict[str, TerraformRequiredProvider] = None
    backend: Union[dict[str, Any], None] = None


class TerraformStack(BaseModel):
    """
    A Terraform stack is terraform-specific flavor of the
    `laktory.models.Stack`. It re-structure the attributes to be aligned with a
     terraform json file.

    It is generally not instantiated directly, but rather created using
    `laktory.models.Stack.to_terraform()`.

    """

    terraform: TerraformConfig = TerraformConfig()
    providers: dict[str, Any] = {}
    resources: dict[str, Any] = {}

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "providers": "provider",
            "resources": "resource",
        }

    @model_validator(mode="after")
    def required_providers(self) -> Any:
        if self.terraform.required_providers is None:
            providers = {}
            for p in self.providers.values():
                providers[p.resource_name_without_alias] = TerraformRequiredProvider(
                    source=p.source, version=p.version
                )
            self.terraform.required_providers = providers

        return self

    def model_dump(self, *args, **kwargs) -> dict[str, Any]:
        """Serialize model to match the structure of a Terraform json file."""
        self._configure_serializer(singular=True)
        kwargs["exclude_none"] = kwargs.get("exclude_none", True)
        d = super().model_dump(*args, **kwargs)

        # Special treatment of resources
        d["resource"] = defaultdict(lambda: {})
        d["data"] = defaultdict(lambda: {})
        for r in self.resources.values():
            _d = r.terraform_properties
            if r.lookup_existing:
                d["data"][r.terraform_resource_lookup_type][r.resource_name] = (
                    r.lookup_existing.model_dump()
                )
                for k in r.options.terraform_options:
                    if k in _d:
                        d["data"][r.terraform_resource_lookup_type][r.resource_name][
                            k
                        ] = _d[k]
            else:
                d["resource"][r.terraform_resource_type][r.resource_name] = _d
        d["data"] = dict(d["data"])
        if len(d["data"]) == 0:
            del d["data"]
        d["resource"] = dict(d["resource"])
        self._configure_serializer(singular=False)

        # Special treatment of moved
        # `moved_from` should generally be used with Terraform, but we also
        # support aliases (Pulumi) for better user experience
        i = -1
        for r in self.resources.values():
            moved_from = r.options.moved_from
            aliases = r.options.aliases
            _from = None
            if moved_from:
                _from = moved_from
            elif aliases:
                _from = aliases[0]
            if _from:
                i += 1
                d[f"moved_{i:05d}"] = {
                    "from": f"{r.terraform_resource_type}.{_from}",
                    "to": f"{r.terraform_resource_type}.{r.resource_name}",
                }

        # Terraform JSON requires the keyword "resources." to be removed and the
        # resource_name to be replaced with resource_type.resource_name.
        _vars = {}
        for r in list(self.resources.values()) + list(self.providers.values()):
            k0 = r.resource_name

            k1 = f"{r.terraform_resource_type}.{r.resource_name}"
            # special treatment for data sources
            if r.lookup_existing:
                k1 = f"data.{r.terraform_resource_lookup_type}.{r.resource_name}"

            if isinstance(r, BaseProvider):
                pattern = r"\$\{resources." + k0 + "}"
                _vars[pattern] = k0

            else:
                # ${resources.resource_name} -> resource_type.resource_name
                pattern = r"\$\{resources\." + k0 + r"\}"
                _vars[pattern] = k1

                # ${resources.resource_name.property} -> ${resource_type.resource_name.property}
                pattern = r"\$\{resources\." + k0 + r"\.(.*?)\}"
                _vars[pattern] = rf"${{{k1}.\1}}"

        # Because all variables are mapped to a string, it is more efficient
        # (>10x) to convert the dict to string before substitution.
        d = json.loads(_resolve_values(json.dumps(d), vars=_vars))

        return d

    # ----------------------------------------------------------------------- #
    # Terraform Methods                                                       #
    # ----------------------------------------------------------------------- #

    def write(self) -> str:
        """
        Write Terraform json configuration file

        Returns
        -------
        :
            Filepath of the configuration file
        """
        filepath = os.path.join(CACHE_ROOT, "stack.tf.json")
        logger.info(f"Writing terraform config at '{filepath}'")

        if not os.path.exists(CACHE_ROOT):
            os.makedirs(CACHE_ROOT)

        text = json.dumps(self.model_dump(), indent=4)

        # Terraform stack file is not a strict format. Some keys might be
        # repeated and require special treatment.

        # Special treatment of providers with aliases
        for _, p in self.providers.items():
            if p.alias is not None:
                text = text.replace(
                    f'"{p.resource_name}":',
                    f'"{p.resource_name_without_alias}":',
                )

        # Special treatment of moved
        text = re.sub(r'"moved_\d+": {', '"moved": {', text)

        # Output
        with open(filepath, "w") as fp:
            fp.write(text)

        return filepath

    def _call(self, command: str, flags: list[str] = None):
        from laktory.cli._common import Worker

        self.write()
        worker = Worker()

        cmd = ["terraform", command]

        if flags is not None:
            cmd += flags

        # Inject user-agent value for monitoring usage as a Databricks partner
        set_databricks_sdk_upstream()

        logger.info(f"Invoking '{' '.join(cmd)}'")
        worker.run(
            cmd=cmd,
            cwd=CACHE_ROOT,
            raise_exceptions=settings.cli_raise_external_exceptions,
        )

    def init(self, flags: list[str] = None) -> None:
        """
        Runs `terraform init`

        Parameters
        ----------
        flags:
            List of flags / options for pulumi plan
        """
        self._call("init", flags=flags)

    def plan(self, flags: list[str] = None) -> None:
        """
        Runs `terraform plan`

        Parameters
        ----------
        flags:
            List of flags / options for pulumi plan
        """
        self._call("plan", flags=flags)

    def apply(self, flags: list[str] = None):
        """
        Runs `terraform apply`

        Parameters
        ----------
        flags:
            List of flags / options for terraform apply
        """
        self._call("apply", flags=flags)

    def destroy(self, flags: list[str] = None):
        """
        Runs `terraform destroy`

        Parameters
        ----------
        flags:
            List of flags / options for terraform destroy
        """
        self._call("destroy", flags=flags)



from typing import Literal
from typing import Union

from pydantic import Field

from laktory._logger import get_logger
from laktory.models.basemodel import BaseModel
from laktory.models.grants.tablegrant import TableGrant
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource

logger = get_logger(__name__)


class TableLookup(ResourceLookup):
    """
    Attributes
    ----------
    name:
        Full name of the databricks_table: `catalog`.`schema`.`table`
    """

    name: str = Field(serialization_alias="id")


class TableColumn(BaseModel):
    """
    A table column.

    Attributes
    ----------
    name:
        User-visible name of column
    comment:
        User-supplied free-form text.
    identity:
        Whether field is an identity column. Can be `default`, `always` or
        `unset`. It is `unset` by default.
    nullable:
        Whether field is nullable (Default: `true`)
    type:
        Column type spec (with metadata) as SQL text. Not supported for `VIEW`
        table_type.
    type_json:

    """

    name: str
    comment: str = None
    identity: str = None
    nullable: bool = None
    type: str = None
    type_json: str = None


class Table(BaseModel, PulumiResource, TerraformResource):
    """
    A table resides in the third layer of Unity Catalog’s three-level
    namespace. It contains rows of data.

    Attributes
    ----------
    catalog_name:
        Name of the catalog storing the table
    columns:
        List of columns stored in the table
    comment:
        Text description of the catalog
    data_source_format:
        External tables are supported in multiple data source formats. The string constants identifying these formats
        are DELTA, CSV, JSON, AVRO, PARQUET, ORC, TEXT. Change forces creation of a new resource. Not supported for
        MANAGED tables or VIEW.
    grant:
        Grant(s) operating on the Table and authoritative for a specific principal.
        Other principals within the grants are preserved. Mutually exclusive with
        `grants`.
    grants:
        Grants operating on the Table and authoritative for all principals.
        Replaces any existing grants defined inside or outside of Laktory. Mutually
        exclusive with `grant`.
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    name:
        Name of the table
    schema_name:
        Name of the schema storing the table
    storage_credential_name:
        For EXTERNAL Tables only: the name of storage credential to use. Change forces creation of a new resource.
    storage_location:
        URL of storage location for Table data (required for EXTERNAL Tables). Not supported for VIEW or MANAGED
        table_type.
    table_type:
        Distinguishes a view vs. managed/external Table.
    view_definition:
        SQL text defining the view (for `table_type == "VIEW"`). Not supported
        for MANAGED or EXTERNAL table_type.
    warehouse_id:
        All table CRUD operations must be executed on a running cluster or SQL
        warehouse. If a warehouse_id is specified, that SQL warehouse will be
        used to execute SQL commands to manage this table.

    Examples
    --------
    ```py
    from laktory import models

    table = models.resources.databricks.Table(
        name="slv_stock_prices",
        table_type="MANAGED",
    )
    ```

    References
    ----------

    * [Databricks Unity Table](https://docs.databricks.com/en/data-governance/unity-catalog/index.html#tables)
    * [Pulumi Databricks Table](https://www.pulumi.com/registry/packages/databricks/api-docs/sqltable/)
    """

    catalog_name: Union[str, None] = None
    columns: Union[list[TableColumn], None] = None
    comment: Union[str, None] = None
    data_source_format: str = "DELTA"
    grant: Union[TableGrant, list[TableGrant]] = None
    grants: list[TableGrant] = None
    lookup_existing: TableLookup = Field(None, exclude=True)
    name: str
    properties: Union[dict[str, str], None] = None
    schema_name: Union[str, None] = None
    storage_credential_name: Union[str, None] = None
    storage_location: Union[str, None] = None
    table_type: Literal["MANAGED", "EXTERNAL", "VIEW"] = "MANAGED"  # required
    view_definition: Union[str, None] = None
    warehouse_id: Union[str, None] = None

    # ----------------------------------------------------------------------- #
    # Validators                                                              #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        """Schema full name `{catalog_name}.{schema_name}`"""
        _id = ""
        if self.catalog_name:
            _id += self.catalog_name

        if self.schema_name:
            if _id == "":
                _id = self.schema_name
            else:
                _id += f".{self.schema_name}"

        return _id

    @property
    def full_name(self) -> str:
        """Table full name `{catalog_name}.{schema_name}.{table_name}`"""

        if self.lookup_existing:
            return self.lookup_existing.name

        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    @property
    def column_names(self) -> list[str]:
        """List of column names"""
        return [c.name for c in self.columns]

    @property
    def is_from_cdc(self) -> bool:
        """If `True` CDC source is used to build the table"""
        return self.builder.is_from_cdc

    # ----------------------------------------------------------------------- #
    #  Methods                                                                #
    # ----------------------------------------------------------------------- #

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """Table full name (catalog.schema.table)"""
        return self.full_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - table grants
        """
        resources = []

        # Table grants
        resources += self.get_grants_additional_resources()

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:SqlTable"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return [
            "grant",
            "grants",
        ]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "columns": "column",
        }

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_sql_table"

    @property
    def terraform_resource_lookup_type(self) -> str:
        return "databricks_table"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Literal
from typing import Union

from pydantic import Field
from pydantic import model_validator

from laktory.models.basemodel import BaseModel
from laktory.models.grants.cataloggrant import CatalogGrant
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.schema import Schema
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class CatalogLookup(ResourceLookup):
    """
    Attributes
    ----------
    name:
        Catalog name
    """

    name: str = Field(serialization_alias="id")


class Catalog(BaseModel, PulumiResource, TerraformResource):
    """
    A catalog is the first layer of Unity Catalog’s three-level namespace. It’s
    used to organize your data assets.

    Attributes
    ----------
    comment:
        Text description of the catalog
    force_destroy:
        If `True` catalog can be deleted, even when not empty
    grant:
        Grant(s) operating on the Catalog and authoritative for a specific principal.
        Other principals within the grants are preserved. Mutually exclusive with
        `grants`.
    grants:
        Grants operating on the Catalog and authoritative for all principals.
        Replaces any existing grants defined inside or outside of Laktory. Mutually
        exclusive with `grant`.
    isolation_mode:
        Whether the catalog is accessible from all workspaces or a specific set
        of workspaces. Can be ISOLATED or OPEN. Setting the catalog to ISOLATED
        will automatically allow access from the current workspace.
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    name:
        Name of the catalog
    owner:
        User/group/service principal name of the catalog owner
    schemas:
        List of schemas stored in the catalog
    storage_root:
        Managed location of the catalog. Location in cloud storage where data
        for managed tables will be stored. If not specified, the location will
        default to the metastore root location.

    Examples
    --------
    ```py
    from laktory import models

    catalog = models.resources.databricks.Catalog(
        name="dev",
        grants=[
            {"principal": "account users", "privileges": ["USE_CATALOG", "USE_SCHEMA"]}
        ],
        schemas=[
            {
                "name": "engineering",
                "grants": [
                    {"principal": "domain-engineering", "privileges": ["SELECT"]}
                ],
            },
            {
                "name": "sources",
                "volumes": [
                    {
                        "name": "landing",
                        "volume_type": "EXTERNAL",
                        "grants": [
                            {
                                "principal": "account users",
                                "privileges": ["READ_VOLUME"],
                            },
                            {
                                "principal": "role-metastore-admins",
                                "privileges": ["WRITE_VOLUME"],
                            },
                        ],
                    },
                ],
            },
        ],
    )
    ```

    References
    ----------

    * [Databricks Unity Catalog](https://docs.databricks.com/en/data-governance/unity-catalog/index.html#catalogs)
    * [Pulumi Databricks Catalog](https://www.pulumi.com/registry/packages/databricks/api-docs/catalog/)
    """

    comment: Union[str, None] = None
    force_destroy: bool = True
    grant: Union[CatalogGrant, list[CatalogGrant]] = None
    grants: list[CatalogGrant] = None
    isolation_mode: Literal["OPEN", "ISOLATED"] = "OPEN"
    lookup_existing: CatalogLookup = Field(None, exclude=True)
    name: str
    owner: str = None
    schemas: list[Schema] = []
    storage_root: str = None

    @model_validator(mode="after")
    def assign_name(self):
        for schema in self.schemas:
            schema.catalog_name = self.name
            schema.assign_name()

        return self

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def full_name(self) -> str:
        """Full name of the catalog `{catalog_name}`"""
        return self.name

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - catalog grants
        - schemas resources
        """
        resources = []

        # Catalog grants
        resources += self.get_grants_additional_resources()

        # Catalog schemas
        if self.schemas:
            for s in self.schemas:
                resources += s.core_resources

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Catalog"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["schemas", "is_unity", "grants", "grant"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_catalog"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Union

from pydantic import Field

from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.groupmember import GroupMember
from laktory.models.resources.databricks.mwspermissionassignment import (
    MwsPermissionAssignment,
)
from laktory.models.resources.databricks.userrole import UserRole
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class UserLookup(ResourceLookup):
    """
    Attributes
    ----------
    user_id:
         ID of the user
    """

    user_id: str = Field(serialization_alias="id")


class User(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks user

    Attributes
    ----------
    acl_principal_id:
        Identifier for use in databricks_access_control_rule_set, e.g. `groups/Some Group`.
    active:
        Either user is active or not. True by default, but can be set to false in case
        of user deactivation with preserving user assets.
    allow_cluster_create:
        This is a field to allow the group to have cluster create privileges. More
        fine grained permissions could be assigned with databricks.Permissions and
        cluster_id argument. Everyone without `allow_cluster_create` argument set, but
        with permission to use Cluster Policy would be able to create clusters, but
        within boundaries of that specific policy.
    allow_instance_pool_create:
        This is a field to allow the group to have instance pool create privileges. More
        fine grained permissions could be assigned with databricks.Permissions and
        instance_pool_id argument.
    databricks_sql_access:
        This is a field to allow the group to have access to Databricks SQL feature in
        User Interface and through databricks_sql_endpoint.
    disable_as_user_deletion:
        Deactivate the user when deleting the resource, rather than deleting the user
        entirely. Defaults to true when the provider is configured at the account-level
        and false when configured at the workspace-level. This flag is exclusive to
        force_delete_repos and force_delete_home_dir flags.
    display_name:
        Display name for the user
    external_id:
        ID of the user in an external identity provider.
    force:
        Ignore `cannot create group: User with username X already exists.` errors and
        implicitly import the specific group into IaC state, enforcing entitlements
        defined in the instance of resource. This functionality is experimental and is
        designed to simplify corner cases, like Azure Active Directory synchronisation.
    force_delete_home_dir:
        This flag determines whether the user's home directory is deleted when the user
        is deleted. It will have not impact when in the accounts SCIM API. False by
        default.
    force_delete_repos:
        This flag determines whether the user's repo directory is deleted when the user
        is deleted. It will have no impact when in the accounts SCIM API. False by
        default.
    group_ids:
        List of the group ids that the user should be member of.
    home:
        Home folder of the user, e.g. /Users/mr.foo@example.com.
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    repos:
        Personal Repos location of the user, e.g. /Repos/mr.foo@example.com.
    roles:
        List of roles assigned to the user e.g. ("account_admin")
    workspace_access
        When `True`, the user is allowed to have workspace access

    Examples
    --------
    ```py
    from laktory import models

    u = models.resources.databricks.User(
        user_name="john.doe@okube.ai",
        display_name="John Doe",
        group_ids=[
            "${resources.group-role-engineer.id}",
            "${resources.group-domain-finance.id}",
        ],
        roles=["account_admin"],
    )
    ```
    """

    acl_principal_id: str = None
    active: bool = None
    allow_cluster_create: bool = None
    allow_instance_pool_create: bool = None
    databricks_sql_access: bool = None
    disable_as_user_deletion: bool = False
    display_name: str = None
    external_id: str = None
    force: bool = None
    force_delete_home_dir: bool = None
    force_delete_repos: bool = None
    group_ids: list[str] = []
    home: str = None
    lookup_existing: UserLookup = Field(None, exclude=True)
    repos: str = None
    roles: list[str] = []
    user_name: str
    workspace_access: bool = None
    workspace_permission_assignments: list[MwsPermissionAssignment] = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.user_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - user roles
        - user group members
        """
        resources = []
        for role in self.roles:
            resources += [
                UserRole(
                    user_id=f"${{resources.{self.resource_name}.id}}",
                    role=role,
                )
            ]

        # Group Member
        for group_id in self.group_ids:
            resources += [
                GroupMember(
                    group_id=group_id,
                    member_id=f"${{resources.{self.resource_name}.id}}",
                )
            ]

        # Workspace Permission Assignments
        if self.workspace_permission_assignments:
            for a in self.workspace_permission_assignments:
                if a.principal_id is None:
                    a.principal_id = f"${{resources.{self.resource_name}.id}}"
                resources += [a]
        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:User"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["groups", "roles", "group_ids", "workspace_permission_assignments"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_user"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import re
from typing import Any
from typing import Literal
from typing import get_args
from typing import get_origin

from pydantic import AliasChoices
from pydantic import BaseModel as _BaseModel
from pydantic import Field
from pydantic import model_validator

from laktory.models.basemodel import BaseModel


class ResourceOptions(BaseModel):
    """
    Resource options for deployment.

    Attributes
    ----------
    is_enabled:
        If `False`, resource is not passed to the IaC backend and is not
        deployed. May be used for deploying resources to specific stack
        environments only or for disabling resources when debugging.
    depends_on:
        Explicit list of resources dependencies.
        Supported by both pulumi and terraform.
    provider:
        Explicit declaration of resources provider.
        Supported by both pulumi and terraform.
    aliases:
        Specify aliases for this resource, so that renaming or refactoring
        doesn’t replace it.
        Pulumi only.
    delete_before_replace:
        Override the default create-before-delete behavior when replacing a
        resource.
        Pulumi only.
    ignore_changes:
        Declare that changes to certain properties should be ignored during a
        diff.
        Pulumi only.
    import_:
        Bring an existing cloud resource into Pulumi.
        Pulumi only.
    parent:
        Establish a parent/child relationship between resources.
        Pulumi only.
    replace_on_changes:
        Declare that changes to certain properties should be treated as forcing
        a replacement.
        Pulumi only.
    """

    # laktory
    is_enabled: bool = True

    # pulumi + terraform
    depends_on: list[str] = []
    provider: str = None
    ignore_changes: list[str] = None

    # pulumi only
    aliases: list[str] = None
    delete_before_replace: bool = True
    import_: str = None
    parent: str = None
    replace_on_changes: list[str] = None

    # terraform only
    moved_from: str = None

    @property
    def pulumi_options(self) -> list[str]:
        return [
            "depends_on",
            "provider",
            "ignore_changes",
            "aliases",
            "delete_before_replace",
            "import_",
            "parent",
            "replace_on_changes",
        ]

    @property
    def terraform_options(self) -> list[str]:
        return [
            "depends_on",
            "provider",
            "ignore_changes",
            "moved_from",
        ]


class ResourceLookup(BaseModel):
    # model_config = ConfigDict(populate_by_name=True)
    def pulumi_dump(self, *args, **kwargs):
        kwargs["by_alias"] = kwargs.get("by_alias", True)
        kwargs["exclude_unset"] = kwargs.get("exclude_unset", True)
        return self.model_dump(*args, **kwargs)


class BaseResource(_BaseModel):
    """
    Parent class for all Laktory models deployable as one or multiple cloud
    core resources. This `BaseResource` class is derived from
    `pydantic.BaseModel`.

    Attributes
    ----------
    resource_name:
        Name of the resource in the context of infrastructure as code. If None,
        `default_resource_name` will be used instead.
    options:
        Resources options specifications
    """

    resource_name_: str = Field(
        None,
        validation_alias=AliasChoices("resource_name_", "resource_name"),
        exclude=True,
    )
    options: ResourceOptions = Field(ResourceOptions(), exclude=True)
    lookup_existing: ResourceLookup = Field(None, exclude=True, frozen=False)
    _core_resources: list[Any] = None

    @model_validator(mode="before")
    @classmethod
    def base_lookup(cls, data: Any) -> Any:
        if data is None:
            return data

        lookup_existing = data.get("lookup_existing", None)
        if not lookup_existing:
            return data

        for fname, f in cls.model_fields.items():
            if f.is_required():
                # Since all field type hints include `var`, we need to isolate
                # intended type hint
                ann = get_args(f.annotation)[0]
                origin = get_origin(ann)
                args = get_args(ann)

                if ann == str:  # noqa: E721
                    data[fname] = ""
                elif origin == Literal:
                    data[fname] = args[0]
                elif str(f.annotation).startswith("list"):
                    data[fname] = []

        for k, v in cls.lookup_defaults().items():
            data[k] = v

        return data

    @model_validator(mode="after")
    def grants_validator(self) -> Any:
        grant = getattr(self, "grant", None)
        grants = getattr(self, "grants", None)
        if grants and grant:
            raise ValueError(
                "`grants` and `grant` are mutually exclusive. Only set one of them."
            )
        return self

    @classmethod
    def lookup_defaults(cls) -> dict:
        return {}

    @property
    def resource_name(self) -> str:
        pattern = re.compile(r"^[a-zA-Z][a-zA-Z0-9-_]*$")

        name = self.default_resource_name
        if self.resource_name_:
            name = self.resource_name_

        if not pattern.match(name):
            raise ValueError(
                f"Resource name `{name}` is invalid. A name must start with a letter or underscore and may contain only letters, digits, underscores, and dashes."
            )

        return name

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    def get_grants_additional_resources(self, options=None):
        from laktory.models.resources.databricks.grant import Grant
        from laktory.models.resources.databricks.grants import Grants

        resources = []
        if options is None:
            options = {}

        if self.grants:
            resources += Grants(
                resource_name=f"grants-{self.resource_name}",
                schema=f"${{resources.{self.resource_name}.id}}",
                grants=[
                    {"principal": g.principal, "privileges": g.privileges}
                    for g in self.grants
                ],
                options=options,
            ).core_resources

        if self.grant:
            grant = self.grant
            if not isinstance(grant, list):
                grant = [grant]
            for g in grant:
                resources += Grant(
                    resource_name=f"grant-{self.resource_name}-{g.principal}",
                    schema=f"${{resources.{self.resource_name}.id}}",
                    principal=g.principal,
                    privileges=g.privileges,
                    options=options,
                ).core_resources

        return resources

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def resource_type_id(self) -> str:
        """
        Resource type id used to build default resource name. Equivalent to
        class name converted to kebab case. e.g.: SecretScope -> secret-scope
        """
        _id = type(self).__name__
        _id = re.sub(
            r"(?<!^)(?=[A-Z])", "-", _id
        ).lower()  # Convert CamelCase to kebab-case
        return _id

    @property
    def resource_key(self) -> str:
        """
        Resource key used to build default resource name. Equivalent to
        name properties if available. Otherwise, empty string.
        """
        return getattr(self, "name", "")

    @property
    def default_resource_name(self) -> str:
        """
        Resource default name constructed as
        - `{self.resource_type_id}-{self.resource_key}`
        - removing ${resources....} tags
        - removing ${vars....} tags
        - Replacing special characters with - to avoid conflicts with resource properties
        """

        if self.resource_type_id not in self.resource_key:
            name = f"{self.resource_type_id}-{self.resource_key}"
        else:
            name = f"{self.resource_key}"

        if name.endswith("-"):
            name = name[:-1]

        # ${resources.x.property} -> x
        pattern = r"\$\{resources\.(.*?)\.(.*?)\}"
        name = re.sub(pattern, r"\1", name)

        # ${vars.x} -> x
        pattern = r"\$\{vars\.(.*?)\}"
        name = re.sub(pattern, r"\1", name)

        # Replace special characters
        chars = [".", "@", "{", "}", "[", "]", "$", "|"]
        for c in chars:
            name = name.replace(c, "-")

        return name

    @property
    def self_as_core_resources(self):
        """Flag set to `True` if self must be included in core resources"""
        return True

    @property
    def additional_core_resources(self):
        return []

    @property
    def core_resources(self):
        """
        List of core resources to be deployed with this laktory model:
        - class instance (self)
        """
        if self._core_resources is None:
            # Add self
            self._core_resources = []
            if self.self_as_core_resources and self.options.is_enabled:
                self._core_resources += [self]

            # Add additional
            def get_additional_resources(r):
                resources = []

                provider = r.options.provider
                k0 = f"${{resources.{r.resource_name}}}"

                for _r in r.additional_core_resources:
                    if not (r.options.is_enabled and _r.options.is_enabled):
                        continue

                    if provider:
                        if _r.options.provider is None:
                            _r.options.provider = provider

                    do = _r.options.depends_on
                    if r.self_as_core_resources and k0 not in do:
                        do += [k0]
                    _r.options.depends_on = do

                    if _r.self_as_core_resources:
                        resources += [_r]

                    for __r in get_additional_resources(_r):
                        resources += [__r]

                return resources

            self._core_resources += get_additional_resources(self)

        return self._core_resources



class MissingColumnError(Exception):
    """"""

    def __init__(self, message="Column is missing", column_name=None):
        if column_name:
            message = f"Column {column_name} is missing"
        super().__init__(message)


class MissingColumnsError(Exception):
    def __init__(self, message="All columns are missing", column_names=None):
        if column_names:
            message = f"Column {column_names} are missing"
        super().__init__(message)


class DataQualityCheckFailedError(Exception):
    def __init__(self, expectation, node=None):
        message = f"Expectation '{expectation.name}' failed"
        if node:
            message += f" on node '{node.name}'"
        message += f" | {expectation.log_msg}"
        super().__init__(message)


class DataQualityExpectationsNotSupported(Exception):
    def __init__(self, expectation, node=None):
        message = (
            f"Expectation '{expectation.name}' not supported for streaming DataFrame"
        )
        if node:
            message += f" on node '{node.name}'"
        message += (
            ". Only ROW type expectations with 0 absolute tolerances are allowed."
        )
        super().__init__(message)



from typing import Union

from .basedatasource import BaseDataSource
from .filedatasource import FileDataSource
from .memorydatasource import MemoryDataSource
from .pipelinenodedatasource import PipelineNodeDataSource
from .tabledatasource import TableDataSource

classes = [
    FileDataSource,
    MemoryDataSource,
    PipelineNodeDataSource,
    TableDataSource,
]

DataSourcesUnion = Union[
    FileDataSource,
    MemoryDataSource,
    PipelineNodeDataSource,
    TableDataSource,
]



from typing import Union

from pydantic_core import CoreSchema
from pydantic_core import core_schema

from laktory.polars import PolarsExpr
from laktory.polars import PolarsLazyFrame
from laktory.spark import SparkColumn
from laktory.spark import SparkDataFrame

AnyDataFrame = Union[SparkDataFrame, PolarsLazyFrame]
"""DataFrame type from any of the supported backend"""

AnyDataFrameColumn = Union[SparkColumn, PolarsExpr]
"""DataFrame column from any of the supported backend"""


class VariableType(str):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type, handler: callable
    ) -> CoreSchema:
        return core_schema.str_schema()


var = VariableType
"""Laktory variable or expression (string)"""

# ResolvableBool: TypeAlias = Union[bool, var]
# """Boolean or laktory variable that can be resolved as a boolean"""
#
# ResolvableFloat: TypeAlias = Union[float, var]
# """Float or laktory variable that can be resolved as a float"""
#
# ResolvableInt: TypeAlias = Union[int, var]
# """Int or laktory variable that can be resolved as an int"""
#
# ResolvableString: TypeAlias = Union[str, var]
# """String or laktory variable that can be resolved as a string"""



from typing import Union

from laktory.models.basemodel import BaseModel
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class MlflowModelTag(BaseModel):
    """
    MLflow model Tags Specifications

    Attributes
    ----------
    key:

    value:

    """

    key: str
    value: str


class MLflowModel(BaseModel, PulumiResource, TerraformResource):
    """
    MLflow Model

    Attributes
    ----------
    access_controls:
        MLflow model access controls
    description:
        The description of the MLflow model.
    name:
        Name of MLflow model. Change of name triggers new resource.
    tags:
        Tags for the MLflow model.

    Examples
    --------
    ```py
    from laktory import models

    mlmodel = models.resources.databricks.MLflowModel(
        name="My MLflow Model",
        description="My MLflow model description",
        tags=[
            {"key": "key1", "value": "value1"},
            {"key": "key2", "value": "value2"},
        ],
        access_controls=[
            {
                "group_name": "account users",
                "permission_level": "CAN_MANAGE_PRODUCTION_VERSIONS",
            }
        ],
    )
    ```
    """

    access_controls: list[AccessControl] = []
    description: str = None
    name: str
    tags: list[MlflowModelTag] = None

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        return self.name.replace("/", "_").replace("\\", "_").replace(" ", "_")

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    registered_model_id=f"${{resources.{self.resource_name}.registered_model_id}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:MlflowModel"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def singularizations(self) -> dict[str, str]:
        return {
            "tags": "tags",
        }

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_mlflow_model"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Union

from pydantic import model_validator

from laktory.models.basemodel import BaseModel
from laktory.models.grants.schemagrant import SchemaGrant
from laktory.models.resources.databricks.table import Table
from laktory.models.resources.databricks.volume import Volume
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class Schema(BaseModel, PulumiResource, TerraformResource):
    """
    A schema (also called a database) is the second layer of Unity Catalog’s
    three-level namespace. A schema organizes tables and views.

    Attributes
    ----------
    catalog_name:
        Name of the catalog storing the schema
    comment:
        Text description of the catalog
    force_destroy:
        If `True` catalog can be deleted, even when not empty
    grant:
        Grant(s) operating on the Schema and authoritative for a specific principal.
        Other principals within the grants are preserved. Mutually exclusive with
        `grants`.
    grants:
        Grants operating on the Schema and authoritative for all principals.
        Replaces any existing grants defined inside or outside of Laktory. Mutually
        exclusive with `grant`.
    isolation_mode:
        Whether the catalog is accessible from all workspaces or a specific set
        of workspaces. Can be ISOLATED or OPEN. Setting the catalog to ISOLATED
        will automatically allow access from the current workspace.
    name:
        Name of the catalog
    storage_root:
        Managed location of the catalog. Location in cloud storage where data
        for managed tables will be stored. If not specified, the location will
        default to the metastore root location.
    tables:
        List of tables stored in the schema
    volumes:
        List of volumes stored in the schema

    Examples
    --------
    ```py
    from laktory import models

    schema = models.resources.databricks.Schema(
        catalog_name="dev",
        name="engineering",
        grants=[{"principal": "domain-engineering", "privileges": ["SELECT"]}],
    )
    ```

    References
    ----------

    * [Databricks Unity Schema](https://docs.databricks.com/en/data-governance/unity-catalog/index.html#schemas)
    * [Pulumi Databricks Schema](https://www.pulumi.com/registry/packages/databricks/api-docs/schema/)
    """

    catalog_name: Union[str, None] = None
    comment: Union[str, None] = None
    force_destroy: bool = True
    grant: Union[SchemaGrant, list[SchemaGrant]] = None
    grants: list[SchemaGrant] = None
    name: str
    storage_root: str = None
    tables: list[Table] = []
    volumes: list[Volume] = []

    @model_validator(mode="after")
    def assign_name(self):
        for table in self.tables:
            table.catalog_name = self.catalog_name
            table.schema_name = self.name
        for volume in self.volumes:
            if self.catalog_name:
                volume.catalog_name = self.catalog_name
            volume.schema_name = self.name

        return self

    # ----------------------------------------------------------------------- #
    # Computed fields                                                         #
    # ----------------------------------------------------------------------- #

    @property
    def parent_full_name(self) -> str:
        """Catalog full name `{catalog_name}`"""
        return self.catalog_name

    @property
    def full_name(self) -> str:
        """Schema full name `{catalog_name}.{schema_name}`"""
        _id = self.name
        if self.parent_full_name is not None:
            _id = f"{self.parent_full_name}.{_id}"
        return _id

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """Schema full name (catalog.schema)"""
        return self.full_name

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - schema grants
        - tables
        - volumes
        """
        resources = []

        # Schema grants
        resources += self.get_grants_additional_resources()

        if self.volumes:
            for v in self.volumes:
                resources += v.core_resources

        if self.tables:
            for t in self.tables:
                resources += t.core_resources

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Schema"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["tables", "volumes", "grant", "grants"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_schema"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



import os
from datetime import datetime
from typing import Any
from typing import Literal
from typing import Union
from zoneinfo import ZoneInfo

from pydantic import ConfigDict
from pydantic import Field

from laktory._logger import get_logger
from laktory._settings import settings
from laktory.models.basemodel import BaseModel
from laktory.models.dataproducer import DataProducer

logger = get_logger(__name__)

# Metadata to exclude
EXCLUDES = [
    "events_root_",
    "event_root_",
    "tstamp_col",
    "tstamp_in_path",
]


class DataEvent(BaseModel):
    """
    Data Event class defines both the context (metadata) describing a data
    event and its content. It is generally used to write data to a storage
    location.

    Attributes
    ----------
    data:
        Event data stored as a (key, value) object
    description:
        Data event description
    event_root:
        Root path for specific event. Default value: `{settings.workspace_landing_root}/events/my-event`
    events_root:
        Root path for all events. Default value: `{settings.workspace_landing_root}/events/`
    name:
        Data event name
    producer:
        Data event producer
    tstamp_col:
        Column storing event UTC timestamp
    tstamp_in_path:
        If `True`, includes timestamp if data event filepath

    Examples
    --------
    This is example one
    ```py
    from datetime import datetime

    from laktory import models

    event = models.DataEvent(
        name="stock_price",
        producer={"name": "yahoo-finance"},
    )
    print(event)
    '''
    variables={} data=None description=None events_root_=None event_root_=None name='stock_price' producer=DataProducer(variables={}, name='yahoo-finance', description=None, party=1) tstamp_col='created_at' tstamp_in_path=True
    '''

    print(event.event_root)
    # > /Volumes/dev/sources/landing/events/yahoo-finance/stock_price/

    event = models.DataEvent(
        name="stock_price",
        producer={"name": "yahoo-finance"},
        data={
            "created_at": datetime(2023, 8, 23),
            "symbol": "GOOGL",
            "open": 130.25,
            "close": 132.33,
        },
    )
    print(event)
    '''
    variables={} data={'created_at': datetime.datetime(2023, 8, 23, 0, 0), 'symbol': 'GOOGL', 'open': 130.25, 'close': 132.33, '_name': 'stock_price', '_producer_name': 'yahoo-finance', '_created_at': datetime.datetime(2023, 8, 23, 0, 0, tzinfo=zoneinfo.ZoneInfo(key='UTC'))} description=None events_root_=None event_root_=None name='stock_price' producer=DataProducer(variables={}, name='yahoo-finance', description=None, party=1) tstamp_col='created_at' tstamp_in_path=True
    '''

    print(event.dirpath)
    # > /Volumes/dev/sources/landing/events/yahoo-finance/stock_price/2023/08/23/

    print(event.get_landing_filepath())
    '''
    /Volumes/dev/sources/landing/events/yahoo-finance/stock_price/2023/08/23/stock_price_20230823T000000000Z.json
    '''

    print(event.get_storage_filepath())
    # > /events/yahoo-finance/stock_price/2023/08/23/stock_price_20230823T000000000Z.json
    ```
    """

    model_config = ConfigDict(extra="forbid", populate_by_name=True)
    data: Union[dict, None] = None
    description: Union[str, None] = None
    events_root_: Union[str, None] = Field(None, alias="events_root")
    event_root_: Union[str, None] = Field(None, alias="event_root")
    name: str
    producer: DataProducer = None
    tstamp_col: str = "created_at"
    tstamp_in_path: bool = True

    def model_post_init(self, __context):
        super().model_post_init(__context)

        # Add metadata
        if self.data is not None:
            self.data["_name"] = self.name
            self.data["_producer_name"] = self.producer.name
            tstamp = self.data.get(self.tstamp_col)
            if isinstance(tstamp, str):
                tstamp = datetime.fromisoformat(tstamp)
            elif tstamp is None:
                tstamp = datetime.utcnow()
            if not tstamp.tzinfo:
                tstamp = tstamp.replace(tzinfo=ZoneInfo("UTC"))
            self.data["_created_at"] = tstamp

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def created_at(self) -> datetime:
        """Timestamp at which event was created."""
        return self.data["_created_at"]

    # ----------------------------------------------------------------------- #
    # Paths                                                                   #
    # ----------------------------------------------------------------------- #

    @property
    def events_root(self) -> str:
        """Must be computed to dynamically account for settings (env variable at run time)"""
        if self.events_root_:
            return self.events_root_
        return settings.workspace_landing_root + "events/"

    @property
    def event_root(self) -> str:
        """
        Root path for the event. Default path is `{self.events_roots}/{producer_name}/{event_name}/`, but it may
        be overwritten by self.event_root_.

        Returns
        -------
        str
            Event path
        """
        if self.event_root_:
            return self.event_root_

        producer = ""
        if self.producer is not None:
            producer = self.producer.name + "/"
        v = f"{self.events_root}{producer}{self.name}/"
        return v

    @property
    def dirpath(self) -> str:
        """Path of the directory storing the event data."""
        dirpath = self.event_root
        if self.tstamp_in_path:
            t = self.created_at
            dirpath += f"{t.year:04d}/{t.month:02d}/{t.day:02d}/"
        return dirpath

    def get_filename(self, fmt: str = "json", suffix: str = None) -> str:
        """
        Get file name associated with the data of the even, given a format `fmt`
        and a `suffix`.

        Parameters
        ----------
        fmt
            File format
        suffix
            File path suffix

        Returns
        -------
        str
            Data file path
        """
        filename = self.name
        if suffix is not None:
            filename += f"_{suffix}"
        if self.tstamp_in_path:
            t = self.created_at
            const = {
                "mus": {"s": 1e-6},
                "s": {"ms": 1000},
            }  # TODO: replace with constants
            total_ms = int(
                (t.second + t.microsecond * const["mus"]["s"]) * const["s"]["ms"]
            )
            time_str = f"{t.hour:02d}{t.minute:02d}{total_ms:05d}Z"
            filename += f"_{t.year:04d}{t.month:02d}{t.day:02d}T{time_str}"

        filename += f".{fmt}"

        return filename

    def get_landing_filepath(self, fmt="json", suffix=None):
        """
        Get file path on the landing mount/volume associated with the data of
        the event, given a format `fmt` and a `suffix`.

        Parameters
        ----------
        fmt
            File format
        suffix
            File path suffix

        Returns
        -------
        str
            Data file path
        """
        return os.path.join(self.dirpath, self.get_filename(fmt, suffix))

    def get_storage_filepath(self, fmt="json", suffix=None):
        """
        Get file path on the landing storage associated with the data of the
        event, given a format `fmt` and a `suffix`.

        Parameters
        ----------
        fmt
            File format
        suffix
            File path suffix

        Returns
        -------
        str
            Data file path
        """
        path = self.get_landing_filepath(fmt=fmt, suffix=suffix)
        path = path.replace(settings.workspace_landing_root, "/")
        return path

    # ----------------------------------------------------------------------- #
    # Output                                                                  #
    # ----------------------------------------------------------------------- #

    def model_dump(self, *args, **kwargs) -> dict[str, Any]:
        kwargs["exclude"] = kwargs.pop("exclude", EXCLUDES)
        kwargs["by_alias"] = kwargs.pop("by_alias", True)
        kwargs["mode"] = kwargs.pop("mode", "json")
        return super().model_dump(*args, **kwargs)

    def model_dump_json(self, *args, **kwargs) -> str:
        kwargs["exclude"] = kwargs.pop("exclude", EXCLUDES)
        return super().model_dump_json(*args, **kwargs)

    def _overwrite_or_skip(
        self, f: callable, path: str, exists: bool, overwrite: bool, skip: bool
    ):
        if exists:
            if skip:
                logger.info(f"Event {self.name} ({path}) already exists. Skipping.")

            else:
                if overwrite:
                    logger.info(f"Writing event {self.name} to {path}")
                    f()
                else:
                    raise FileExistsError(f"Object {path} already exists.")

        else:
            logger.info(f"Writing event {self.name} to {path}")
            f()

    def to_path(
        self,
        suffix: str = None,
        fmt: str = "json",
        overwrite: bool = False,
        skip_if_exists: bool = False,
    ) -> None:
        """
        Write data event to local file path, given a format `fmt` and a
        `suffix`.

        Parameters
        ----------
        suffix:
            File path suffix
        fmt:
            File format
        overwrite:
            Overwrite file if already exists
        skip_if_exists:
            If `True` and file already exists, writing is skipped.
        """
        path = self.get_landing_filepath(suffix=suffix, fmt=fmt)
        dirpath = os.path.dirname(path)

        if not os.path.exists(dirpath):
            os.makedirs(dirpath)

        def _write():
            if fmt != "json":
                raise NotImplementedError()

            with open(path, "w") as fp:
                fp.write(self.model_dump_json())

        self._overwrite_or_skip(
            f=_write,
            path=path,
            exists=os.path.exists(path),
            overwrite=overwrite,
            skip=skip_if_exists,
        )

    def to_azure_storage_container(
        self,
        suffix: str = None,
        fmt: str = "json",
        container_client: Any = None,
        account_url: str = None,
        container_name: str = "landing",
        overwrite: bool = False,
        skip_if_exists: bool = False,
    ) -> None:
        """
        Write data event to azure storage container, given a format `fmt` and a
        `suffix`.

        Parameters
        ----------
        suffix:
            File path suffix
        fmt:
            File format
        container_client:
            Authorized Azure container client
        account_url:
            URL of storage account
        container_name:
            Name of the storage container
        overwrite:
            Overwrite file if already exists
        skip_if_exists:
            If `True` and file already exists, writing is skipped.
        """
        # Set container client
        if container_client is None:
            from azure.identity import DefaultAzureCredential
            from azure.storage.blob import ContainerClient

            if account_url:
                # From account URL
                # TODO: test
                container_client = ContainerClient(
                    account_url,
                    credential=DefaultAzureCredential(),
                    container_name=container_name,
                )

            elif settings.lakehouse_sa_conn_str is not None:
                # From connection string
                container_client = ContainerClient.from_connection_string(
                    conn_str=settings.lakehouse_sa_conn_str,
                    container_name=container_name,
                )

            else:
                raise ValueError(
                    "Provide a valid container client, an account url or a connection string in LAKEHOUSE_SA_CONN_STR"
                )

        path = self.get_storage_filepath(suffix=suffix, fmt=fmt)
        blob = container_client.get_blob_client(path)

        def _write():
            if fmt != "json":
                raise NotImplementedError()

            blob.upload_blob(
                self.model_dump_json(),
                overwrite=overwrite,
            )

        self._overwrite_or_skip(
            f=_write,
            path=path,
            exists=blob.exists(),
            overwrite=overwrite,
            skip=skip_if_exists,
        )

    def to_aws_s3_bucket(
        self,
        bucket_name,
        suffix: str = None,
        fmt: str = "json",
        s3_resource: Any = None,
        overwrite: bool = False,
        skip_if_exists: bool = False,
    ) -> None:
        """
        Write data event to azure storage container, given a format `fmt` and a
        `suffix`.

        Parameters
        ----------
        suffix:
            File path suffix
        fmt:
            File format
        s3_resource:
            Authorized S3 bucket resource
        overwrite:
            Overwrite file if already exists
        skip_if_exists:
            If `True` and file already exists, writing is skipped.
        """
        import boto3

        if s3_resource is None:
            s3_resource = boto3.resource(
                service_name="s3",
                aws_access_key_id=settings.aws_access_key_id,
                aws_secret_access_key=settings.aws_secret_access_key,
                region_name=settings.aws_region,
            )

        bucket = s3_resource.Bucket(bucket_name)

        path = self.get_storage_filepath(suffix=suffix, fmt=fmt)[1:]

        object_exists = True
        try:
            o = bucket.Object(path)
            o.load()
        except Exception as e:
            if "404" in str(e):
                object_exists = False
            else:
                raise e

        def _write():
            if fmt != "json":
                raise NotImplementedError()

            bucket.put_object(
                Key=path,
                Body=self.model_dump_json(),
            )

        self._overwrite_or_skip(
            f=_write,
            path=path,
            exists=object_exists,
            overwrite=overwrite,
            skip=skip_if_exists,
        )

    def to_gcp_storage_bucket(self):
        raise NotImplementedError()

    def to_databricks(
        self,
        suffix: str = None,
        fmt: str = "json",
        overwrite: bool = False,
        skip_if_exists: bool = False,
        storage_type: Literal["VOLUME", "MOUNT"] = "VOLUME",
    ) -> None:
        """
        Write data event to Databricks volume or mount, given a format `fmt`
        and a `suffix`.

        Parameters
        ----------
        suffix:
            File path suffix
        fmt:
            File format
        overwrite:
            Overwrite file if already exists
        skip_if_exists:
            If `True` and file already exists, writing is skipped.
        storage_type:
            Specify if data is written to a mount or a volume.
        """
        path = self.get_landing_filepath(suffix=suffix, fmt=fmt)
        if storage_type == "MOUNT":
            path = "/dbfs" + path
        dirpath = os.path.dirname(path)

        if not os.path.exists(dirpath):
            os.makedirs(dirpath)

        def _write():
            if fmt != "json":
                raise NotImplementedError()

            with open(path, "w") as fp:
                fp.write(self.model_dump_json())

        self._overwrite_or_skip(
            f=_write,
            path=path,
            exists=os.path.exists(path),
            overwrite=overwrite,
            skip=skip_if_exists,
        )


if __name__ == "__main__":
    from laktory import models

    event = models.DataEvent(
        name="stock_price",
        producer={"name": "yahoo-finance"},
        data={
            "created_at": datetime(2023, 8, 23),
            "symbol": "GOOGL",
            "open": 130.25,
            "close": 132.33,
        },
    )
    print(event)

    print(event.dirpath)
    print(event.get_landing_filepath())
    print(event.get_storage_filepath())



from pathlib import Path


class DataFramesFactory:
    def __init__(self, spark_factory):
        self._brz = None
        self._slv = None
        self._slv_polars = None
        self._slv_stream = None
        self._meta = None
        self._meta_polars = None
        self._name = None
        self._spark = None
        self.data_dirpath = Path(__file__).parent / "../../tests/data/"
        self.spark_factory = spark_factory

    @property
    def spark(self):
        if self._spark is None:
            self._spark = self.spark_factory.spark
        return self._spark

    @property
    def brz(self):
        if self._brz is None:
            self._brz = self.spark.read.parquet(
                str(self.data_dirpath / "brz_stock_prices")
            )
        return self._brz

    @property
    def slv(self):
        if self._slv is None:
            self._slv = self.spark.read.parquet(
                str(self.data_dirpath / "slv_stock_prices")
            )
        return self._slv

    @property
    def slv_stream(self):
        if self._slv_stream is None:
            self._slv_stream = (
                self.spark.readStream.format("delta")
                .option("startingOffsets", "earliest")
                .load(str(self.data_dirpath / "slv_stock_prices_delta"))
            )
        return self._slv_stream

    @property
    def meta(self):
        if self._meta is None:
            self._meta = self.spark.read.parquet(
                str(self.data_dirpath / "slv_stock_meta")
            )
        return self._meta

    @property
    def name(self):
        if self._name is None:
            import pandas as pd

            self._name = self.spark.createDataFrame(
                pd.DataFrame(
                    {
                        "symbol3": ["AAPL", "GOOGL", "AMZN"],
                        "name": ["Apple", "Google", "Amazon"],
                    }
                )
            )
        return self._name

    @property
    def slv_polars(self):
        if self._slv_polars is None:
            import polars as pl

            dirpath = self.data_dirpath / "slv_stock_prices"
            for filename in dirpath.iterdir():
                if str(filename).endswith(".parquet"):
                    break
            self._slv_polars = pl.read_parquet(dirpath / filename)
        return self._slv_polars

    @property
    def meta_polars(self):
        if self._meta_polars is None:
            import polars as pl

            dirpath = self.data_dirpath / "slv_stock_meta"
            for filename in dirpath.iterdir():
                if str(filename).endswith(".parquet"):
                    break
            self._meta_polars = pl.read_parquet(dirpath / filename)
        return self._meta_polars



import os
from pathlib import Path
from typing import Any
from typing import Literal
from typing import Union

from pydantic import Field
from pydantic import model_validator

from laktory import settings
from laktory.models.basemodel import BaseModel
from laktory.models.resources.baseresource import ResourceLookup
from laktory.models.resources.databricks.accesscontrol import AccessControl
from laktory.models.resources.databricks.permissions import Permissions
from laktory.models.resources.pulumiresource import PulumiResource
from laktory.models.resources.terraformresource import TerraformResource


class NotebookLookup(ResourceLookup):
    """
    Attributes
    ----------
    path:
        Notebook path on the workspace
    format:
        Notebook format to export. Either `SOURCE`, `HTML`, `JUPYTER`, or `DBC`
    """

    path: str = Field(serialization_alias="id")
    format: str = "SOURCE"


class Notebook(BaseModel, PulumiResource, TerraformResource):
    """
    Databricks Notebook

    Attributes
    ----------
    access_controls:
        List of notebook access controls
    dirpath:
        Workspace directory inside rootpath in which the notebook is deployed.
        Used only if `path` is not specified.
    language:
         Notebook programming language
    lookup_existing:
        Specifications for looking up existing resource. Other attributes will
        be ignored.
    path:
        Workspace filepath for the notebook. Overwrite `rootpath` and `dirpath`.
    rootpath:
        Root directory to which all notebooks are deployed to. Can also be
        configured by settings LAKTORY_WORKSPACE_LAKTORY_ROOT environment
        variable. Default is `/.laktory/`. Used only if `path` is not
        specified.
    source:
        Path to notebook in source code format on local filesystem.

    Examples
    --------
    ```py
    from laktory import models

    notebook = models.resources.databricks.Notebook(
        source="./notebooks/dlt/dlt_laktory_pl.py",
    )
    print(notebook.path)
    # > /.laktory/dlt_laktory_pl.py

    notebook = models.resources.databricks.Notebook(
        source="./notebooks/dlt/dlt_laktory_pl.py",
        rootpath="/src/",
    )
    print(notebook.path)
    # > /src/dlt_laktory_pl.py

    notebook = models.resources.databricks.Notebook(
        source="./notebooks/dlt/dlt_laktory_pl.py",
        rootpath="/src/",
        dirpath="notebooks/dlt/",
    )
    print(notebook.path)
    # > /src/notebooks/dlt/dlt_laktory_pl.py
    ```
    """

    access_controls: list[AccessControl] = []
    dirpath: str = None
    language: Literal["SCALA", "PYTHON", "SQL", "R"] = None
    lookup_existing: NotebookLookup = Field(None, exclude=True)
    path: str = None
    rootpath: str = None
    source: str

    @property
    def filename(self) -> str:
        """Notebook file name"""
        return os.path.basename(self.source)

    @model_validator(mode="after")
    def set_paths(self) -> Any:
        # Path set
        if self.path:
            return self

        # root
        if self.rootpath is None:
            self.rootpath = settings.workspace_laktory_root

        # dir
        if self.dirpath is None:
            self.dirpath = ""
        if self.dirpath.startswith("/"):
            self.dirpath = self.dirpath[1:]

        # path
        _path = Path(self.rootpath) / self.dirpath / self.filename
        self.path = _path.as_posix()

        return self

    # ----------------------------------------------------------------------- #
    # Resource Properties                                                     #
    # ----------------------------------------------------------------------- #

    @property
    def resource_key(self) -> str:
        """path with special characters `/`, `.`, `\\` replaced with `-`"""
        key = self.path
        if key is None:
            return ""
        key = key.replace("/", "-")
        key = key.replace("\\", "-")
        key = key.replace(".", "-")
        for i in range(5):
            if key.startswith("-"):
                key = key[1:]
        return key

    @property
    def additional_core_resources(self) -> list[PulumiResource]:
        """
        - permissions
        """
        resources = []
        if self.access_controls:
            resources += [
                Permissions(
                    resource_name=f"permissions-{self.resource_name}",
                    access_controls=self.access_controls,
                    notebook_path=f"${{resources.{self.resource_name}.path}}",
                )
            ]

        return resources

    # ----------------------------------------------------------------------- #
    # Pulumi Properties                                                       #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_resource_type(self) -> str:
        return "databricks:Notebook"

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        return ["access_controls", "dirpath", "rootpath"]

    # ----------------------------------------------------------------------- #
    # Terraform Properties                                                    #
    # ----------------------------------------------------------------------- #

    @property
    def terraform_resource_type(self) -> str:
        return "databricks_notebook"

    @property
    def terraform_excludes(self) -> Union[list[str], dict[str, bool]]:
        return self.pulumi_excludes



from typing import Union

from .basedatasink import BaseDataSink
from .basedatasink import DataSinkMergeCDCOptions
from .filedatasink import FileDataSink
from .tabledatasink import TableDataSink

classes = [
    FileDataSink,
    TableDataSink,
]


DataSinksUnion = Union[
    FileDataSink,
    TableDataSink,
]



from abc import abstractmethod
from typing import Any
from typing import Union

from laktory._parsers import _snake_to_camel
from laktory.models.resources.baseresource import BaseResource


class PulumiResource(BaseResource):
    """
    Parent class for all Laktory models deployable with Pulumi IaC backend.
    """

    _pulumi_resources: dict[str, Any] = {}

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    @abstractmethod
    def pulumi_resource_type(self) -> str:
        raise NotImplementedError()

    @property
    def pulumi_excludes(self) -> Union[list[str], dict[str, bool]]:
        """List of fields to exclude when dumping model to pulumi"""
        return []

    @property
    def pulumi_renames(self) -> dict[str, str]:
        """Map of fields to rename when dumping model to pulumi"""
        return {}

    # ----------------------------------------------------------------------- #
    # Methods                                                                 #
    # ----------------------------------------------------------------------- #

    @property
    def pulumi_properties(self) -> dict:
        """
        Resources properties formatted for pulumi:

        * Serialization (model dump)
        * Removal of excludes defined in `self.pulumi_excludes`
        * Renaming of keys according to `self.pulumi_renames`
        * Injection of variables

        Returns
        -------
        :
            Pulumi-safe model dump
        """
        # `exclude_unset` should be used instead of exclude_none, but for now
        # we set default values to facilitate instantiation of models within
        # other models (permissions in MwsPermissionAssignment for example).
        # A better approach should be used.
        d = super().model_dump(exclude=self.pulumi_excludes, exclude_none=True)
        for k, v in self.pulumi_renames.items():
            if self._camel_serialization:
                k = _snake_to_camel(k)
                v = _snake_to_camel(v)
            if k in d:
                d[v] = d.pop(k)
        return d



import io
import sys

from pyspark.sql.dataframe import DataFrame


def is_aggregate(df: DataFrame) -> bool:
    """
    Check if DataFrame is an aggregation.

    Parameters
    ----------
    df:
        Input DataFrame

    Returns
    -------
    :
        Result

    Examples
    --------

    ```py
    import pyspark.sql.functions as F

    import laktory  # noqa: F401

    data = (
        {"symbol": "AAPL", "close": 1},
        {"symbol": "AAPL", "close": 2},
        {"symbol": "AAPL", "close": 3},
        {"symbol": "AMZN", "close": 4},
        {"symbol": "AMZN", "close": 5},
        {"symbol": "AMZN", "close": 6},
    )

    df = spark.createDataFrame(data)
    dfa = df.groupby("symbol").agg(F.mean("close"))

    dfa.laktory.is_aggregate()
    ```
    """
    buffer = io.StringIO()
    stdout0 = sys.stdout
    sys.stdout = buffer
    df.explain(mode="simple")
    sys.stdout = stdout0
    plan = buffer.getvalue()
    return "HashAggregate".lower() in plan.lower()



import time
from typing import Union

from pyspark.sql.dataframe import DataFrame


def display(
    df: DataFrame,
    n: int = 10,
    truncate: Union[bool, int] = False,
    vertical: bool = False,
    refresh_interval: float = 3.0,
) -> None:
    """
    Prints the first n rows to the console. Compatible for both static and
    streaming dataframes. In the case of a streaming dataframe, console will
    be continuously updated until there is a keyboard input.

    Parameters
    ----------
    df:
        Input DataFrame
    n:
        Number of rows to display
    truncate:
        If set to `True`, truncate strings longer than 20 chars by default. If
        set to a number greater than one, truncates long strings to length
        truncate and align cells right.
    vertical:
        If set to `True`, print output rows vertically (one line per column
        value).
    refresh_interval:
        Pause duration, in seconds, between each update for streaming
        dataframes


    Returns
    -------
    :
        None

    Examples
    --------

    ```py
    import pandas as pd

    import laktory  # noqa: F401

    df = spark.createDataFrame(
        pd.DataFrame(
            {
                "symbol": ["AAPL", "GOOGL"],
                "price": [200.0, 205.0],
                "tstamp": ["2023-09-01", "2023-09-01"],
            }
        )
    )
    df.laktory.display(n=5, refresh_interval=2)
    ```
    """

    if not df.isStreaming:
        df.show(n=n, truncate=truncate, vertical=vertical)

    else:
        # Start the streaming query
        query = (
            df.writeStream.outputMode("append")
            .format("memory")  # Store the results in-memory table
            .queryName("_laktory_tmp_view")
            .start()
        )

        try:
            while True:
                # Fetch and display the latest rows
                df.sparkSession.sql(f"SELECT * FROM _laktory_tmp_view LIMIT {n}").show(
                    truncate=truncate, vertical=vertical
                )
                time.sleep(refresh_interval)
        except KeyboardInterrupt:
            print("Stopped streaming display.")
        finally:
            query.stop()
